# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Phantoms given by simple geometric objects such as cubes or spheres."""

from __future__ import absolute_import, division, print_function

import numpy as np

from odl.core.discr.discr_space import uniform_discr_fromdiscr
from odl.core.util.numerics import resize_array

__all__ = (
    'cuboid',
    'defrise',
    'ellipsoid_phantom',
    'indicate_proj_axis',
    'smooth_cuboid',
    'tgv_phantom',
)


def cuboid(space, min_pt=None, max_pt=None):
    """Rectangular cuboid.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Space in which the phantom should be created.
    min_pt : array-like of shape ``(space.ndim,)``, optional
        Lower left corner of the cuboid. If ``None`` is given, a quarter
        of the extent from ``space.min_pt`` towards the inside is chosen.
    max_pt : array-like of shape ``(space.ndim,)``, optional
        Upper right corner of the cuboid. If ``None`` is given, ``min_pt``
        plus half the extent is chosen.

    Returns
    -------
    phantom : `DiscretizedSpaceElement`
        The generated cuboid phantom in ``space``.

    Examples
    --------
    If both ``min_pt`` and ``max_pt`` are omitted, the cuboid lies in the
    middle of the space domain and extends halfway towards all sides:

    >>> space = odl.uniform_discr([0, 0], [1, 1], [4, 6])
    >>> odl.core.phantom.cuboid(space)
    uniform_discr([ 0.,  0.], [ 1.,  1.], (4, 6)).element(
        [[ 0.,  0.,  0.,  0.,  0.,  0.],
         [ 0.,  1.,  1.,  1.,  1.,  0.],
         [ 0.,  1.,  1.,  1.,  1.,  0.],
         [ 0.,  0.,  0.,  0.,  0.,  0.]]
    )

    By specifying the corners, the cuboid can be arbitrarily placed and
    scaled:

    >>> odl.core.phantom.cuboid(space, [0.25, 0], [0.75, 0.5])
    uniform_discr([ 0.,  0.], [ 1.,  1.], (4, 6)).element(
        [[ 0.,  0.,  0.,  0.,  0.,  0.],
         [ 1.,  1.,  1.,  0.,  0.,  0.],
         [ 1.,  1.,  1.,  0.,  0.,  0.],
         [ 0.,  0.,  0.,  0.,  0.,  0.]]
    )
    """
    dom_min_pt = np.asarray(space.domain.min())
    dom_max_pt = np.asarray(space.domain.max())

    if min_pt is None:
        min_pt = dom_min_pt * 0.75 + dom_max_pt * 0.25
    if max_pt is None:
        max_pt = dom_min_pt * 0.25 + dom_max_pt * 0.75

    min_pt = np.atleast_1d(min_pt)
    max_pt = np.atleast_1d(max_pt)

    if min_pt.shape != (space.ndim,):
        raise ValueError('shape of `min_pt` must be {}, got {}'
                         ''.format((space.ndim,), min_pt.shape))
    if max_pt.shape != (space.ndim,):
        raise ValueError('shape of `max_pt` must be {}, got {}'
                         ''.format((space.ndim,), max_pt.shape))

    def phantom(x):
        result = True
        for xi, xmin, xmax in zip(x, min_pt, max_pt):
            xmin = space.array_backend.array_constructor(
                xmin, device=space.device
            )
            xmax = space.array_backend.array_constructor(
                xmax, device=space.device
            )
            result = (result &
                      space.array_namespace.less_equal(xmin, xi) & 
                      space.array_namespace.less_equal(xi, xmax)
            )
        return result

    return space.element(phantom)


def defrise(space, nellipses=8, alternating=False, min_pt=None, max_pt=None):
    """Phantom with regularily spaced ellipses.

    This phantom is often used to verify cone-beam algorithms.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Space in which the phantom should be created, must be 2- or
        3-dimensional.
    nellipses : int, optional
        Number of ellipses. If more ellipses are used, each ellipse becomes
        thinner.
    alternating : bool, optional
        True if the ellipses should have alternating densities (+1, -1),
        otherwise all ellipses have value +1.
    min_pt, max_pt : array-like, optional
        If provided, use these vectors to determine the bounding box of the
        phantom instead of ``space.min_pt`` and ``space.max_pt``.
        It is currently required that ``min_pt >= space.min_pt`` and
        ``max_pt <= space.max_pt``, i.e., shifting or scaling outside the
        original space is not allowed.

        Providing one of them results in a shift, e.g., for ``min_pt``::

            new_min_pt = min_pt
            new_max_pt = space.max_pt + (min_pt - space.min_pt)

        Providing both results in a scaled version of the phantom.

    Returns
    -------
    phantom : ``space`` element
        The generated phantom in ``space``.

    See Also
    --------
    odl.core.phantom.transmission.shepp_logan
    """
    ellipses = defrise_ellipses(space.ndim, nellipses=nellipses,
                                alternating=alternating)
    return ellipsoid_phantom(space, ellipses, min_pt, max_pt)


def defrise_ellipses(ndim, nellipses=8, alternating=False):
    """Ellipses for the standard Defrise phantom in 2 or 3 dimensions.

    Parameters
    ----------
    ndim : {2, 3}
        Dimension of the space for the ellipses/ellipsoids.
    nellipses : int, optional
        Number of ellipses. If more ellipses are used, each ellipse becomes
        thinner.
    alternating : bool, optional
        True if the ellipses should have alternating densities (+1, -1),
        otherwise all ellipses have value +1.

    See Also
    --------
    odl.core.phantom.geometric.ellipsoid_phantom :
        Function for creating arbitrary ellipsoids phantoms
    odl.core.phantom.transmission.shepp_logan_ellipsoids
    """
    ellipses = []
    if ndim == 2:
        for i in range(nellipses):
            if alternating:
                value = (-1.0 + 2.0 * (i % 2))
            else:
                value = 1.0

            axis_1 = 0.5
            axis_2 = 0.5 / (nellipses + 1)
            center_x = 0.0
            center_y = -1 + 2.0 / (nellipses + 1.0) * (i + 1)
            rotation = 0
            ellipses.append(
                [value, axis_1, axis_2, center_x, center_y, rotation])
    elif ndim == 3:
        for i in range(nellipses):
            if alternating:
                value = (-1.0 + 2.0 * (i % 2))
            else:
                value = 1.0

            axis_1 = axis_2 = 0.5
            axis_3 = 0.5 / (nellipses + 1)
            center_x = center_y = 0.0
            center_z = -1 + 2.0 / (nellipses + 1.0) * (i + 1)
            rotation_phi = rotation_theta = rotation_psi = 0

            ellipses.append(
                [value, axis_1, axis_2, axis_3,
                 center_x, center_y, center_z,
                 rotation_phi, rotation_theta, rotation_psi])

    return ellipses


def indicate_proj_axis(space, scale_structures=0.5):
    """Phantom indicating along which axis it is projected.

    The number (n) of rectangles in a parallel-beam projection along a main
    axis (0, 1, or 2) indicates the projection to be along the (n-1)the
    dimension.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Space in which the phantom should be created, must be 2- or
        3-dimensional.
    scale_structures : positive float in (0, 1], optional
        Scales objects (cube, cuboids)

    Returns
    -------
    phantom : ``space`` element
        Projection helper phantom in ``space``.

    Examples
    --------
    Phantom in 2D space:

    >>> space = odl.uniform_discr([0, 0], [1, 1], shape=(8, 8))
    >>> phantom = indicate_proj_axis(space).asarray()
    >>> print(odl.core.util.array_str(phantom, nprint=10))
    [[ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]

    >>> space = odl.uniform_discr([0] * 3, [1] * 3, [8, 8, 8])
    >>> phantom = odl.core.phantom.indicate_proj_axis(space).asarray()
    >>> axis_sum_0 = np.sum(phantom, axis=0)
    >>> print(odl.core.util.array_str(axis_sum_0, nprint=10))
    [[ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  3.,  3.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  3.,  3.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]
    >>> axis_sum_1 = np.sum(phantom, axis=1)
    >>> print(odl.core.util.array_str(axis_sum_1, nprint=10))
    [[ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  2.,  2.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  2.,  2.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]
    >>> axis_sum_2 = np.sum(phantom, axis=2)
    >>> print(odl.core.util.array_str(axis_sum_2, nprint=10))
    [[ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  2.,  2.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  2.,  2.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  2.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  2.,  0.,  0.,  0.,  0.],
     [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]]
    """
    if not 0 < scale_structures <= 1:
        raise ValueError('`scale_structures` ({}) is not in (0, 1]'
                         ''.format(scale_structures))

    assert space.ndim in (2, 3)

    shape = space.shape
    phan = np.zeros(shape)
    shape = np.array(shape) - 1
    cen = np.round(0.5 * shape)
    dx = np.floor(scale_structures * 0.25 * shape)
    dx[dx == 0] = 1

    # cube of size 2 * dx, offset in x axis, symmetric in others
    ix0 = int((cen - 3 * dx)[0])
    if space.ndim == 2:
        ix, iy = (cen - 1 * dx).astype(int)
        phan[ix0:ix, iy:-iy] = 1
    elif space.ndim == 3:
        ix, iy, iz = (cen - 1 * dx).astype(int)
        phan[ix0:ix, iy:-iy, iz:-iz] = 1

    # 1st cuboid of size (dx[0], dx[1], 2 * dx[2]), offset in x and y axes,
    # symmetric in z axis
    ix0 = int((cen + 1 * dx)[1])
    ix1 = int((cen + 2 * dx)[1])
    iy0 = int(cen[1])
    if space.ndim == 2:
        phan[ix0:ix1, iy0:-iy] = 1
    elif space.ndim == 3:
        iz = int((cen - dx)[2])
        phan[ix0:ix1, iy0:-iy, iz:-iz] = 1

    # 2nd cuboid of (dx[0], dx[1], 2 * dx[2]) touching the first diagonally
    # at a long edge; offset in x and y axes, symmetric in z axis
    ix0 = int((cen + 2 * dx)[1])
    ix1 = int((cen + 3 * dx)[1])
    iy1 = int(cen[1])
    if space.ndim == 2:
        phan[ix0:ix1, iy:iy1] = 1
    elif space.ndim == 3:
        iz = int((cen - dx)[2])
        phan[ix0:ix1, iy:iy1, iz:-iz] = 1

    return space.element(phan)


def _getshapes_2d(center, max_radius, shape):
    """Calculate indices and slices for the bounding box of a disk."""
    index_mean = shape * center
    index_radius = max_radius / 2.0 * np.array(shape)

    # Avoid negative indices
    min_idx = np.maximum(np.floor(index_mean - index_radius), 0).astype(int)
    max_idx = np.ceil(index_mean + index_radius).astype(int)
    idx = [slice(minx, maxx) for minx, maxx in zip(min_idx, max_idx)]
    shapes = [(idx[0], slice(None)),
              (slice(None), idx[1])]
    return tuple(idx), tuple(shapes)


def _ellipse_phantom_2d(space, ellipses):
    """Create a phantom of ellipses in 2d space.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Uniformly discretized space in which the phantom should be generated.
        If ``space.shape`` is 1 in an axis, a corresponding slice of the
        phantom is created (instead of squashing the whole phantom into the
        slice).
    ellipses : list of lists
        Each row should contain the entries ::

            'value',
            'axis_1', 'axis_2',
            'center_x', 'center_y',
            'rotation'

        The provided ellipses need to be specified relative to the
        reference rectangle ``[-1, -1] x [1, 1]``. Angles are to be given
        in radians.

    Returns
    -------
    phantom : ``space`` element
        2D ellipse phantom in ``space``.

    See Also
    --------
    shepp_logan : The typical use-case for this function.
    """
    # Blank image
    p = np.zeros(space.shape, dtype=space.dtype_identifier)

    minp = space.grid.min_pt
    maxp = space.grid.max_pt

    # Create the pixel grid
    grid_in = space.grid.meshgrid

    # move points to [-1, 1]
    grid = []
    for i in range(2):
        mean_i = (minp[i] + maxp[i]) / 2.0
        # Where space.shape = 1, we have minp = maxp, so we set diff_i = 1
        # to avoid division by zero. Effectively, this allows constructing
        # a slice of a 2D phantom.
        diff_i = (maxp[i] - minp[i]) / 2.0 or 1.0
        grid.append((grid_in[i] - mean_i) / diff_i)

    for ellip in ellipses:
        assert len(ellip) == 6

        intensity = ellip[0]
        a_squared = ellip[1] ** 2
        b_squared = ellip[2] ** 2
        x0 = ellip[3]
        y0 = ellip[4]
        theta = ellip[5]

        scales = [1 / a_squared, 1 / b_squared]
        center = (np.array([x0, y0]) + 1.0) / 2.0

        # Create the offset x,y and z values for the grid
        if theta != 0:
            # Rotate the points to the expected coordinate system.
            ctheta = np.cos(theta)
            stheta = np.sin(theta)

            mat = np.array([[ctheta, stheta],
                            [-stheta, ctheta]])

            # Calculate the points that could possibly be inside the volume
            # Since the points are rotated, we cannot do anything directional
            # without more logic
            max_radius = np.sqrt(
                np.abs(mat).dot([a_squared, b_squared]))
            idx, shapes = _getshapes_2d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            offset_points = [vec * (xi - x0i)[..., None]
                             for xi, vec, x0i in zip(subgrid,
                                                     mat.T,
                                                     [x0, y0])]
            rotated = offset_points[0] + offset_points[1]
            np.square(rotated, out=rotated)
            radius = np.dot(rotated, scales)
        else:
            # Calculate the points that could possibly be inside the volume
            max_radius = np.sqrt([a_squared, b_squared])
            idx, shapes = _getshapes_2d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            squared_dist = [ai * (xi - x0i) ** 2
                            for xi, ai, x0i in zip(subgrid,
                                                   scales,
                                                   [x0, y0])]

            # Parentheses to get best order for broadcasting
            radius = squared_dist[0] + squared_dist[1]

        # Find the points within the ellipse
        inside = radius <= 1

        # Add the ellipse intensity to those points
        p[idx][inside] += intensity

    return space.element(p)


def _getshapes_3d(center, max_radius, shape):
    """Calculate indices and slices for the bounding box of a ball."""
    index_mean = shape * center
    index_radius = max_radius / 2.0 * np.array(shape)

    min_idx = np.floor(index_mean - index_radius).astype(int)
    min_idx = np.maximum(min_idx, 0)  # avoid negative indices
    max_idx = np.ceil(index_mean + index_radius).astype(int)
    idx = [slice(minx, maxx) for minx, maxx in zip(min_idx, max_idx)]
    shapes = [(idx[0], slice(None), slice(None)),
              (slice(None), idx[1], slice(None)),
              (slice(None), slice(None), idx[2])]
    return tuple(idx), tuple(shapes)


def _ellipsoid_phantom_3d(space, ellipsoids):
    """Create an ellipsoid phantom in 3d space.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Space in which the phantom should be generated. If ``space.shape`` is
        1 in an axis, a corresponding slice of the phantom is created
        (instead of squashing the whole phantom into the slice).
    ellipsoids : list of lists
        Each row should contain the entries ::

            'value',
            'axis_1', 'axis_2', 'axis_3',
            'center_x', 'center_y', 'center_z',
            'rotation_phi', 'rotation_theta', 'rotation_psi'

        The provided ellipsoids need to be specified relative to the
        reference cube ``[-1, -1, -1] x [1, 1, 1]``. Angles are to be given
        in radians.

    Returns
    -------
    phantom : ``space`` element
        3D ellipsoid phantom in ``space``.

    See Also
    --------
    shepp_logan : The typical use-case for this function.
    """
    # Blank volume
    p = np.zeros(space.shape, dtype=space.dtype_identifier)

    minp = space.grid.min_pt
    maxp = space.grid.max_pt

    # Create the pixel grid
    grid_in = space.grid.meshgrid

    # Move points to [-1, 1]
    grid = []
    for i in range(3):
        mean_i = (minp[i] + maxp[i]) / 2.0
        # Where space.shape = 1, we have minp = maxp, so we set diff_i = 1
        # to avoid division by zero. Effectively, this allows constructing
        # a slice of a 3D phantom.
        diff_i = (maxp[i] - minp[i]) / 2.0 or 1.0
        grid.append((grid_in[i] - mean_i) / diff_i)

    for ellip in ellipsoids:
        assert len(ellip) == 10

        intensity = ellip[0]
        a_squared = ellip[1] ** 2
        b_squared = ellip[2] ** 2
        c_squared = ellip[3] ** 2
        x0 = ellip[4]
        y0 = ellip[5]
        z0 = ellip[6]
        phi = ellip[7]
        theta = ellip[8]
        psi = ellip[9]

        scales = [1 / a_squared, 1 / b_squared, 1 / c_squared]
        center = (np.array([x0, y0, z0]) + 1.0) / 2.0

        # Create the offset x,y and z values for the grid
        if any([phi, theta, psi]):
            # Rotate the points to the expected coordinate system.
            cphi = np.cos(phi)
            sphi = np.sin(phi)
            ctheta = np.cos(theta)
            stheta = np.sin(theta)
            cpsi = np.cos(psi)
            spsi = np.sin(psi)

            mat = np.array([[cpsi * cphi - ctheta * sphi * spsi,
                             cpsi * sphi + ctheta * cphi * spsi,
                             spsi * stheta],
                            [-spsi * cphi - ctheta * sphi * cpsi,
                             -spsi * sphi + ctheta * cphi * cpsi,
                             cpsi * stheta],
                            [stheta * sphi,
                             -stheta * cphi,
                             ctheta]])

            # Calculate the points that could possibly be inside the volume
            # Since the points are rotated, we cannot do anything directional
            # without more logic
            max_radius = np.sqrt(
                np.abs(mat).dot([a_squared, b_squared, c_squared]))
            idx, shapes = _getshapes_3d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            offset_points = [vec * (xi - x0i)[..., None]
                             for xi, vec, x0i in zip(subgrid,
                                                     mat.T,
                                                     [x0, y0, z0])]
            rotated = offset_points[0] + offset_points[1] + offset_points[2]
            np.square(rotated, out=rotated)
            radius = np.dot(rotated, scales)
        else:
            # Calculate the points that could possibly be inside the volume
            max_radius = np.sqrt([a_squared, b_squared, c_squared])
            idx, shapes = _getshapes_3d(center, max_radius, space.shape)

            subgrid = [g[idi] for g, idi in zip(grid, shapes)]
            squared_dist = [ai * (xi - x0i) ** 2
                            for xi, ai, x0i in zip(subgrid,
                                                   scales,
                                                   [x0, y0, z0])]

            # Parentheses to get best order for broadcasting
            radius = squared_dist[0] + (squared_dist[1] + squared_dist[2])

        # Find the points within the ellipse
        inside = radius <= 1

        # Add the ellipse intensity to those points
        p[idx][inside] += intensity

    return space.element(p)


def ellipsoid_phantom(space, ellipsoids, min_pt=None, max_pt=None):
    """Return a phantom given by ellipsoids.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Space in which the phantom should be created, must be 2- or
        3-dimensional. If ``space.shape`` is 1 in an axis, a corresponding
        slice of the phantom is created (instead of squashing the whole
        phantom into the slice).
    ellipsoids : sequence of sequences
        If ``space`` is 2-dimensional, each row should contain the entries ::

            'value',
            'axis_1', 'axis_2',
            'center_x', 'center_y',
            'rotation'

        If ``space`` is 3-dimensional, each row should contain the entries ::

            'value',
            'axis_1', 'axis_2', 'axis_3',
            'center_x', 'center_y', 'center_z',
            'rotation_phi', 'rotation_theta', 'rotation_psi'

        The provided ellipsoids need to be specified relative to the
        reference rectangle ``[-1, -1] x [1, 1]``, or analogously in 3d.
        The angles are to be given in radians.

    min_pt, max_pt : array-like, optional
        If provided, use these vectors to determine the bounding box of the
        phantom instead of ``space.min_pt`` and ``space.max_pt``.
        It is currently required that ``min_pt >= space.min_pt`` and
        ``max_pt <= space.max_pt``, i.e., shifting or scaling outside the
        original space is not allowed.

        Providing one of them results in a shift, e.g., for ``min_pt``::

            new_min_pt = min_pt
            new_max_pt = space.max_pt + (min_pt - space.min_pt)

        Providing both results in a scaled version of the phantom.

    Notes
    -----
    The phantom is created by adding the values of each ellipse. The
    ellipses are defined by a center point
    ``(center_x, center_y, [center_z])``, the lengths of its principial
    axes ``(axis_1, axis_2, [axis_2])``, and a rotation angle ``rotation``
    in 2D or Euler angles ``(rotation_phi, rotation_theta, rotation_psi)``
    in 3D.

    This function is heavily optimized, achieving runtimes about 20 times
    faster than "trivial" implementations. It is therefore recommended to use
    it in all phantoms where applicable.

    The main optimization is that it only considers a subset of all the
    points when updating for each ellipse. It does this by first finding
    a subset of points that could possibly be inside the ellipse. This
    optimization is very good for "spherical" ellipsoids, but not so
    much for elongated or rotated ones.

    It also does calculations wherever possible on the meshgrid instead of
    individual points.

    Examples
    --------
    Create a circle with a smaller circle inside:

    >>> space = odl.uniform_discr([-1, -1], [1, 1], [5, 5])
    >>> ellipses = [[1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
    ...             [1.0, 0.6, 0.6, 0.0, 0.0, 0.0]]
    >>> print(ellipsoid_phantom(space, ellipses))
    [[ 0.,  0.,  1.,  0.,  0.],
     [ 0.,  1.,  2.,  1.,  0.],
     [ 1.,  2.,  2.,  2.,  1.],
     [ 0.,  1.,  2.,  1.,  0.],
     [ 0.,  0.,  1.,  0.,  0.]]

    See Also
    --------
    odl.core.phantom.transmission.shepp_logan : Classical Shepp-Logan phantom,
        typically used for transmission imaging
    odl.core.phantom.transmission.shepp_logan_ellipsoids : Ellipses for the
        Shepp-Logan phantom
    odl.core.phantom.geometric.defrise_ellipses : Ellipses for the
        Defrise phantom
    """
    if space.ndim == 2:
        _phantom = _ellipse_phantom_2d
    elif space.ndim == 3:
        _phantom = _ellipsoid_phantom_3d
    else:
        raise ValueError('dimension not 2 or 3, no phantom available')

    if min_pt is None and max_pt is None:
        return _phantom(space, ellipsoids)

    else:
        # Generate a temporary space with given `min_pt` and `max_pt`
        # (snapped to the cell grid), create the phantom in that space and
        # resize to the target size for `space`.
        # The snapped points are constructed by finding the index of
        # `min/max_pt` in the space partition, indexing the partition with
        # that index, yielding a single-cell partition, and then taking
        # the lower-left/upper-right corner of that cell.
        if min_pt is None:
            snapped_min_pt = space.min_pt
        else:
            min_pt_cell = space.partition[space.partition.index(min_pt)]
            snapped_min_pt = min_pt_cell.min_pt

        if max_pt is None:
            snapped_max_pt = space.max_pt
        else:
            max_pt_cell = space.partition[space.partition.index(max_pt)]
            snapped_max_pt = max_pt_cell.max_pt
            # Avoid snapping to the next cell where max_pt falls exactly on
            # a boundary
            for i in range(space.ndim):
                if max_pt[i] in space.partition.cell_boundary_vecs[i]:
                    snapped_max_pt[i] = max_pt[i]

        tmp_space = uniform_discr_fromdiscr(
            space, min_pt=snapped_min_pt, max_pt=snapped_max_pt,
            cell_sides=space.cell_sides)

        tmp_phantom = _phantom(tmp_space, ellipsoids)
        offset = space.partition.index(tmp_space.min_pt)
        return space.element(
            resize_array(tmp_phantom, space.shape, offset))


def smooth_cuboid(space, min_pt=None, max_pt=None, axis=0):
    """Cuboid with smooth variations.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Discretized space in which the phantom is supposed to be created.
    min_pt : array-like of shape ``(space.ndim,)``, optional
        Lower left corner of the cuboid. If ``None`` is given, a quarter
        of the extent from ``space.min_pt`` towards the inside is chosen.
    max_pt : array-like of shape ``(space.ndim,)``, optional
        Upper right corner of the cuboid. If ``None`` is given, ``min_pt``
        plus half the extent is chosen.
    axis : int or sequence of int
        Dimension(s) along which the smooth variation should happen.

    Returns
    -------
    phantom : ``space``-element
        The generated cuboid phantom in ``space``. Values have range [0, 1].
    """
    dom_min_pt = space.domain.min()
    dom_max_pt = space.domain.max()

    if min_pt is None:
        min_pt = dom_min_pt * 0.75 + dom_max_pt * 0.25
    if max_pt is None:
        max_pt = dom_min_pt * 0.25 + dom_max_pt * 0.75

    min_pt = np.atleast_1d(min_pt)
    max_pt = np.atleast_1d(max_pt)

    axis = np.array(axis, dtype=int, ndmin=1)

    if min_pt.shape != (space.ndim,):
        raise ValueError('shape of `min_pt` must be {}, got {}'
                         ''.format((space.ndim,), min_pt.shape))
    if max_pt.shape != (space.ndim,):
        raise ValueError('shape of `max_pt` must be {}, got {}'
                         ''.format((space.ndim,), max_pt.shape))

    sign = 0
    for i, coord in enumerate(space.meshgrid):
        sign = sign | (coord < min_pt[i]) | (coord > max_pt[i])

    values = 0
    for i in axis:
        coord = space.meshgrid[i]
        extent = (dom_max_pt[i] - dom_min_pt[i])
        values = values + 2 * (coord - dom_min_pt[i]) / extent - 1

    # Properly scale using sign
    sign = (3 * sign - 2) / axis.size

    # Fit in [0, 1]
    values = values * sign
    values = (values - np.min(values)) / (np.max(values) - np.min(values))

    return space.element(values)


def tgv_phantom(space, edge_smoothing=0.2):
    """Piecewise affine phantom.

    This phantom is taken from [Bre+2010] and includes both linearly varying
    regions and sharp discontinuities. It is designed to work well with
    Total Generalized Variation (TGV) type regularization.

    Parameters
    ----------
    space : `DiscretizedSpace`, 2 dimensional
        Discretized space in which the phantom is supposed to be created.
        Needs to be two-dimensional.
    edge_smoothing : nonnegative float, optional
        Smoothing of the edges of the phantom, given as smoothing width in
        units of minimum pixel size.

    Returns
    -------
    phantom : ``space``-element
        The generated phantom in ``space``. Values have range [0, 1].

    Notes
    -----
    The original phantom is given by a specific image. In this implementation,
    we extracted the underlying parameters and the phantom thus works with
    spaces of any shape. Due to this, small variations may occur when compared
    to the original phantom.

    References
    ----------
    [Bre+2010] K. Bredies, K. Kunisch, and T. Pock.
    *Total Generalized Variation*. SIAM Journal on Imaging Sciences,
    3(3):492-526, Jan. 2010
    """
    if space.ndim != 2:
        raise ValueError('`space.ndim` must be 2, got {}'
                         ''.format(space.ndim))

    y, x = space.meshgrid

    # Use a smooth sigmoid to get some anti-aliasing across edges.
    scale = edge_smoothing / np.min(space.shape)

    def sigmoid(val):
        if edge_smoothing != 0:
            val = val / scale
            with np.errstate(over="ignore", under="ignore"):
                return 1 / (1 + np.exp(-val))
        else:
            return (val > 0).astype(val.dtype)

    # Normalize to [0, 1]
    x = (x - np.min(x)) / (np.max(x) - np.min(x))
    y = (y - np.min(y)) / (np.max(y) - np.min(y))

    # Background
    values = -(x + y) / 2

    # Square-ish region
    indicator = np.ones(space.shape)
    indicator *= sigmoid(-(0.015199034981905914 * x - y + 0.13896260554885403))
    indicator *= sigmoid((0.3333333333333323 * y - x + 0.598958333333334))
    indicator *= sigmoid((-2.4193548387096726 * y - x + 2.684979838709672))

    values += indicator * 2 * (x + y - 1)

    # Ellipse part
    x_c = x - 0.71606842360499456
    y_c = y - 0.18357884949910641

    width = 0.55677657235995637
    height = 0.37279391542283741
    phi = 0.62911754900697558

    x_c_rot = (np.cos(phi) * x_c - np.sin(phi) * y_c) / width
    y_c_rot = (np.sin(phi) * x_c + np.cos(phi) * y_c) / height

    indicator = sigmoid(np.sqrt(x_c_rot ** 2 + y_c_rot ** 2) - 1)

    values = indicator * values + 1.5 * (1 - indicator) * (-x - 2 * y + 0.6)

    # Normalize values
    values = (values - np.min(values)) / (np.max(values) - np.min(values))

    return space.element(values)


if __name__ == '__main__':
    # Show the phantoms
    import odl

    # cuboid 1D
    space = odl.uniform_discr(-1, 1, 300)
    cuboid(space).show('cuboid 1d')

    # cuboid 2D
    space = odl.uniform_discr([-1, -1], [1, 1], [300, 300])
    cuboid(space).show('cuboid 2d')

    # smooth cuboid
    smooth_cuboid(space).show('smooth_cuboid x 2d')
    smooth_cuboid(space, axis=[0, 1]).show('smooth_cuboid x-y 2d')

    # TGV phantom
    tgv_phantom(space).show('tgv_phantom')

    # cuboid 3D
    space = odl.uniform_discr([-1, -1, -1], [1, 1, 1], [300, 300, 300])
    cuboid(space).show('cuboid 3d')

    # Indicate proj axis 3D
    indicate_proj_axis(space).show('indicate_proj_axis 3d')

    # ellipsoid phantom 2D
    space = odl.uniform_discr([-1, -1], [1, 1], [300, 300])
    ellipses = [[1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.6, 0.6, 0.0, 0.0, 0.0]]
    ellipsoid_phantom(space, ellipses).show('ellipse phantom 2d')

    # ellipsoid phantom 3D
    space = odl.uniform_discr([-1, -1, -1], [1, 1, 1], [300, 300, 300])
    ellipsoids = [[1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                  [1.0, 0.6, 0.6, 0.6, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]
    ellipsoid_phantom(space, ellipsoids).show('ellipsoid phantom 3d')

    # Defrise phantom 2D
    space = odl.uniform_discr([-1, -1], [1, 1], [300, 300])
    defrise(space).show('defrise 2D')

    # Defrise phantom 2D
    space = odl.uniform_discr([-1, -1, -1], [1, 1, 1], [300, 300, 300])
    defrise(space).show('defrise 3D', coords=[0, None, None])

    # Run also the doctests
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""(Quasi-)Newton schemes to find zeros of functionals."""

from __future__ import print_function, division, absolute_import
import numpy as np

from odl.solvers.util import ConstantLineSearch
from odl.solvers.iterative.iterative import conjugate_gradient


__all__ = ('newtons_method', 'bfgs_method', 'broydens_method')


def _bfgs_direction(s, y, x, hessinv_estimate=None):
    r"""Compute ``Hn^-1(x)`` for the L-BFGS method.

    Parameters
    ----------
    s : sequence of `LinearSpaceElement`
        The ``s`` coefficients in the BFGS update, see Notes.
    y : sequence of `LinearSpaceElement`
        The ``y`` coefficients in the BFGS update, see Notes.
    x : `LinearSpaceElement`
        Point in which to evaluate the product.
    hessinv_estimate : `Operator`, optional
        Initial estimate of the hessian ``H0^-1``.

    Returns
    -------
    r : ``x.space`` element
        The result of ``Hn^-1(x)``.

    Notes
    -----
    :math:`H_n^{-1}` is defined recursively as

    .. math::
        H_{n+1}^{-1} =
        \left(I - \frac{ s_n y_n^T}{y_n^T s_n} \right)
        H_{n}^{-1}
        \left(I - \frac{ y_n s_n^T}{y_n^T s_n} \right) +
        \frac{s_n s_n^T}{y_n^T \, s_n}

    With :math:`H_0^{-1}` given by ``hess_estimate``.
    """
    assert len(s) == len(y)

    r = x.copy()
    alphas = np.zeros(len(s))
    rhos = np.zeros(len(s))

    for i in reversed(range(len(s))):
        rhos[i] = 1.0 / y[i].inner(s[i])
        alphas[i] = rhos[i] * (s[i].inner(r))
        r.lincomb(1, r, -alphas[i], y[i])

    if hessinv_estimate is not None:
        r = hessinv_estimate(r)

    for i in range(len(s)):
        beta = rhos[i] * (y[i].inner(r))
        r.lincomb(1, r, alphas[i] - beta, s[i])

    return r


def _broydens_direction(s, y, x, hessinv_estimate=None, impl='first'):
    r"""Compute ``Hn^-1(x)`` for Broydens method.

    Parameters
    ----------
    s : sequence of `LinearSpaceElement`'s'
        The ``s`` coefficients in the Broydens update, see Notes.
    y : sequence of `LinearSpaceElement`'s'
        The ``y`` coefficients in the Broydens update, see Notes.
    x : `LinearSpaceElement`
        Point in which to evaluate the product.
    hessinv_estimate : `Operator`, optional
        Initial estimate of the hessian ``H0^-1``.
    impl : {'first', 'second'}, optional
        The type of Broydens method to use.

    Returns
    -------
    r : ``x.space`` element
        The result of ``Hn^-1(x)``.

    Notes
    -----
    For ``impl = 'first'``, :math:`H_n^{-1}` is defined recursively as

    .. math::
        H_{n+1}^{-1} = \left(I + s_n y_n^T \right) H_{n}^{-1}

    and for ``impl = 'second'``:

    .. math::
        H_{n+1}^{-1} = H_{n}^{-1} + s_n y_n^T

    With :math:`H_0^{-1}` given by ``hess_estimate``.
    """
    assert len(s) == len(y)

    if hessinv_estimate is not None:
        r = hessinv_estimate(x)
    else:
        r = x.copy()

    for i in range(len(s)):
        if impl == 'first':
            r.lincomb(1, r, y[i].inner(r), s[i])
        elif impl == 'second':
            r.lincomb(1, r, y[i].inner(x), s[i])
        else:
            raise RuntimeError('unknown `impl`')

    return r


def newtons_method(f, x, line_search=1.0, maxiter=1000, tol=1e-16,
                   cg_iter=None, callback=None):
    r"""Newton's method for minimizing a functional.

    Notes
    -----
    This is a general and optimized implementation of Newton's method
    for solving the problem:

    .. math::
        \min f(x)

    for a differentiable function
    :math:`f: \mathcal{X}\to \mathbb{R}` on a Hilbert space
    :math:`\mathcal{X}`. It does so by finding a zero of the gradient

    .. math::
        \nabla f: \mathcal{X} \to \mathcal{X}.

    of finding a root of a function.

    The algorithm is well-known and there is a vast literature about it.
    Among others, the method is described in [BV2004], Sections 9.5
    and 10.2 (`book available online
    <http://stanford.edu/~boyd/cvxbook/bv_cvxbook.pdf>`_),
    [GNS2009],  Section 2.7 for solving nonlinear equations and Section
    11.3 for its use in minimization, and wikipedia on `Newton's_method
    <https://en.wikipedia.org/wiki/Newton's_method>`_.

    The algorithm works by iteratively solving

    .. math::
        \partial f(x_k)p_k = -f(x_k)

    and then updating as

    .. math::
        x_{k+1} = x_k + \alpha x_k,

    where :math:`\alpha` is a suitable step length (see the
    references). In this implementation the system of equations are
    solved using the conjugate gradient method.

    Parameters
    ----------
    f : `Functional`
        Goal functional. Needs to have ``f.gradient`` and
        ``f.gradient.derivative``.
    x : ``op.domain`` element
        Starting point of the iteration
    line_search : float or `LineSearch`, optional
        Strategy to choose the step length. If a float is given, uses it as a
        fixed step length.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Tolerance that should be used for terminating the iteration.
    cg_iter : int, optional
        Number of iterations in the the conjugate gradient solver,
        for computing the search direction.
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate

    References
    ----------
    [BV2004] Boyd, S, and Vandenberghe, L. *Convex optimization*.
    Cambridge university press, 2004.

    [GNS2009] Griva, I, Nash, S G, and Sofer, A. *Linear and nonlinear
    optimization*. Siam, 2009.
    """
    # TODO: update doc
    grad = f.gradient
    if x not in grad.domain:
        raise TypeError('`x` {!r} is not in the domain of `f` {!r}'
                        ''.format(x, grad.domain))

    if not callable(line_search):
        line_search = ConstantLineSearch(line_search)

    if cg_iter is None:
        # Motivated by that if it is Ax = b, x and b in Rn, it takes at most n
        # iterations to solve with cg
        cg_iter = grad.domain.size

    # TODO: optimize by using lincomb and avoiding to create copies
    for _ in range(maxiter):

        # Initialize the search direction to 0
        search_direction = x.space.zero()

        # Compute hessian (as operator) and gradient in the current point
        hessian = grad.derivative(x)
        deriv_in_point = grad(x)

        # Solving A*x = b for x, in this case f''(x)*p = -f'(x)
        # TODO: Let the user provide/choose method for how to solve this?
        try:
            hessian_inverse = hessian.inverse
        except NotImplementedError:
            conjugate_gradient(hessian, search_direction,
                               -deriv_in_point, cg_iter)
        else:
            hessian_inverse(-deriv_in_point, out=search_direction)

        # Computing step length
        dir_deriv = search_direction.inner(deriv_in_point)
        if np.abs(dir_deriv) <= tol:
            return

        step_length = line_search(x, search_direction, dir_deriv)

        # Updating
        x += step_length * search_direction

        if callback is not None:
            callback(x)


def bfgs_method(f, x, line_search=1.0, maxiter=1000, tol=1e-15, num_store=None,
                hessinv_estimate=None, callback=None):
    r"""Quasi-Newton BFGS method to minimize a differentiable function.

    Notes
    -----
    Can use either the regular BFGS method, or the limited memory BFGS method.

    This is a general and optimized implementation of a quasi-Newton
    method with BFGS update for solving a general unconstrained
    optimization problem

    .. math::
        \min f(x)

    for a differentiable function
    :math:`f: \mathcal{X}\to \mathbb{R}` on a Hilbert space
    :math:`\mathcal{X}`. It does so by finding a zero of the gradient

    .. math::
        \nabla f: \mathcal{X} \to \mathcal{X}.

    The QN method is an approximate Newton method, where the Hessian
    is approximated and gradually updated in each step. This
    implementation uses the rank-one BFGS update schema where the
    inverse of the Hessian is recalculated in each iteration.

    The algorithm is described in [GNS2009], Section 12.3 and in the
    `BFGS Wikipedia article
    <https://en.wikipedia.org/wiki/BFGS_algorithm>`_.

    Parameters
    ----------
    f : `Functional`
        Functional with ``f.gradient``.
    x : ``f.domain`` element
        Starting point of the iteration
    line_search : float or `LineSearch`, optional
        Strategy to choose the step length. If a float is given, uses it as a
        fixed step length.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Tolerance that should be used for terminating the iteration.
    num_store : int, optional
        Maximum number of correction factors to store. For ``None``, the method
        is the regular BFGS method. For an integer, the method becomes the
        Limited Memory BFGS method.
    hessinv_estimate : `Operator`, optional
        Initial estimate of the inverse of the Hessian operator. Needs to be an
        operator from ``f.domain`` to ``f.domain``.
        Default: Identity on ``f.domain``
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate.

    References
    ----------
    [GNS2009] Griva, I, Nash, S G, and Sofer, A. *Linear and nonlinear
    optimization*. Siam, 2009.
    """
    grad = f.gradient
    if x not in grad.domain:
        raise TypeError('`x` {!r} is not in the domain of `grad` {!r}'
                        ''.format(x, grad.domain))

    if not callable(line_search):
        line_search = ConstantLineSearch(line_search)

    ys = []
    ss = []

    grad_x = grad(x)
    for i in range(maxiter):
        # Determine a stepsize using line search
        search_dir = -_bfgs_direction(ss, ys, grad_x, hessinv_estimate)
        dir_deriv = search_dir.inner(grad_x)
        if np.abs(dir_deriv) == 0:
            return  # we found an optimum
        step = line_search(x, direction=search_dir, dir_derivative=dir_deriv)

        # Update x
        x_update = search_dir
        x_update *= step
        x += x_update

        grad_x, grad_diff = grad(x), grad_x
        # grad_diff = grad(x) - grad(x_old)
        grad_diff.lincomb(-1, grad_diff, 1, grad_x)

        y_inner_s = grad_diff.inner(x_update)

        # Test for convergence
        if np.abs(y_inner_s) < tol:
            if grad_x.norm() < tol:
                return
            else:
                # Reset if needed
                ys = []
                ss = []
                continue

        # Update Hessian
        ys.append(grad_diff)
        ss.append(x_update)
        if num_store is not None:
            # Throw away factors if they are too many.
            ss = ss[-num_store:]
            ys = ys[-num_store:]

        if callback is not None:
            callback(x)


def broydens_method(f, x, line_search=1.0, impl='first', maxiter=1000,
                    tol=1e-15, hessinv_estimate=None,
                    callback=None):
    r"""Broyden's first method, a quasi-Newton scheme.

    Notes
    -----
    This is a general and optimized implementation of Broyden's method,
    a quasi-Newton method for solving a general unconstrained optimization
    problem

    .. math::
        \min f(x)

    for a differentiable function
    :math:`f: \mathcal{X}\to \mathbb{R}` on a Hilbert space
    :math:`\mathcal{X}`. It does so by finding a zero of the gradient

    .. math::
        \nabla f: \mathcal{X} \to \mathcal{X}

    using a Newton-type update scheme with approximate Hessian.

    The algorithm is described in [Bro1965] and [Kva1991], and in a
    `Wikipedia article
    <https://en.wikipedia.org/wiki/Broyden's_method>`_.

    Parameters
    ----------
    f : `Functional`
        Functional with ``f.gradient``
    x : ``f.domain`` element
        Starting point of the iteration
    line_search : float or `LineSearch`, optional
        Strategy to choose the step length. If a float is given, uses it as a
        fixed step length.
    impl : {'first', 'second'}, optional
        What version of Broydens method to use. First is also known as Broydens
        'good' method, while the second is known as Broydens 'bad' method.
    maxiter : int, optional
        Maximum number of iterations.
        ``tol``.
    tol : float, optional
        Tolerance that should be used for terminating the iteration.
    hessinv_estimate : `Operator`, optional
        Initial estimate of the inverse of the Hessian operator. Needs to be an
        operator from ``f.domain`` to ``f.domain``.
        Default: Identity on ``f.domain``
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate.

    References
    ----------
    [Bro1965] Broyden, C G. *A class of methods for solving nonlinear
    simultaneous equations*. Mathematics of computation, 33 (1965),
    pp 577--593.

    [Kva1991] Kvaalen, E. *A faster Broyden method*. BIT Numerical
    Mathematics 31 (1991), pp 369--372.
    """
    grad = f.gradient
    if x not in grad.domain:
        raise TypeError('`x` {!r} is not in the domain of `grad` {!r}'
                        ''.format(x, grad.domain))

    if not callable(line_search):
        line_search = ConstantLineSearch(line_search)

    impl, impl_in = str(impl).lower(), impl
    if impl not in ('first', 'second'):
        raise ValueError('`impl` {!r} not understood'
                         ''.format(impl_in))

    ss = []
    ys = []

    grad_x = grad(x)
    for i in range(maxiter):
        # find step size
        search_dir = -_broydens_direction(ss, ys, grad_x,
                                          hessinv_estimate, impl)
        dir_deriv = search_dir.inner(grad_x)
        if np.abs(dir_deriv) == 0:
            return  # we found an optimum

        step = line_search(x, search_dir, dir_deriv)

        # update x
        x_update = step * search_dir
        x += x_update

        # compute new gradient
        grad_x, grad_x_old = grad(x), grad_x
        delta_grad = grad_x - grad_x_old

        # update hessian.
        # TODO: reuse from above
        v = _broydens_direction(ss, ys, delta_grad, hessinv_estimate,
                                impl)
        if impl == 'first':
            divisor = x_update.inner(v)

            # Test for convergence
            if np.abs(divisor) < tol:
                if grad_x.norm() < tol:
                    return
                else:
                    # Reset if needed
                    ys = []
                    ss = []
                    continue
            u = (x_update - v) / divisor
            ss.append(u)
            ys.append(x_update)
        elif impl == 'second':
            divisor = delta_grad.inner(delta_grad)

            # Test for convergence
            if np.abs(divisor) < tol:
                if grad_x.norm() < tol:
                    return
                else:
                    # Reset if needed
                    ys = []
                    ss = []
                    continue
            u = (x_update - v) / divisor
            ss.append(u)
            ys.append(delta_grad)

        if callback is not None:
            callback(x)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities for internal functionality connected to vectorization."""

from __future__ import print_function, division, absolute_import
from builtins import object
from functools import wraps
import numpy as np

from odl.core.array_API_support import get_array_and_backend


__all__ = ('is_valid_input_array', 'is_valid_input_meshgrid',
           'out_shape_from_meshgrid', 'out_shape_from_array',
           'OptionalArgDecorator', 'vectorize')


def is_valid_input_array(x, ndim=None):
    """Test if ``x`` is a correctly shaped point array in R^d."""
    try:
        x, backend = get_array_and_backend(x)
    except ValueError:
        # raising a ValueError here will be problematic when cheking lists/tuple.
        if isinstance(x, (list, tuple)):
            x = np.asarray(x)
        else:
            return False

    if ndim is None or ndim == 1:
        return x.ndim == 1 and x.size > 1 or x.ndim == 2 and x.shape[0] == 1
    else:
        return x.ndim == 2 and x.shape[0] == ndim


def is_valid_input_meshgrid(x, ndim):
    """Test if ``x`` is a `meshgrid` sequence for points in R^d."""
    # This case is triggered in FunctionSpaceElement.__call__ if the
    # domain does not have an 'ndim' attribute. We return False and
    # continue.
    if ndim is None:
        return False

    if not isinstance(x, tuple):
        return False

    if ndim > 1:
        try:
            np.broadcast(*x)
        except (ValueError, TypeError):  # cannot be broadcast
            return False

    return (len(x) == ndim and
            all(isinstance(xi, np.ndarray) for xi in x) and
            all(xi.ndim == ndim for xi in x))


def out_shape_from_meshgrid(mesh):
    """Get the broadcast output shape from a `meshgrid`."""
    if len(mesh) == 1:
        return (len(mesh[0]),)
    else:
        # Ragged arrays are a liability in the current implementation
        _, backend = get_array_and_backend(mesh[0])
        namespace = backend.array_namespace
        if backend.impl == 'numpy':
            return namespace.broadcast(*mesh).shape
        elif backend.impl == 'pytorch':
            mesh_size = namespace.broadcast_shapes(
                    *(t.shape for t in mesh))
            return list(mesh_size)
        else:
            raise NotImplementedError(f'Not implemented for impl {backend.impl}')   
              
def out_shape_from_array(arr):
    """Get the output shape from an array."""
    if isinstance(arr, (float, int, complex, list, tuple)):
        arr = np.asarray(arr)
    else:
        arr,_ = get_array_and_backend(arr)
    if arr.ndim == 1:
        return arr.shape
    else:
        return (arr.shape[1],)


class OptionalArgDecorator(object):

    """Abstract class to create decorators with optional arguments.

    This class implements the functionality of a decorator that can
    be used with and without arguments, i.e. the following patterns
    both work::

        @decorator
        def myfunc(x, *args, **kwargs):
            pass

        @decorator(param, **dec_kwargs)
        def myfunc(x, *args, **kwargs):
            pass

    The arguments to the decorator are passed on to the underlying
    wrapper.

    To use this class, subclass it and implement the static ``_wrapper``
    method.
    """

    def __new__(cls, *args, **kwargs):
        """Create a new decorator instance.

        There are two cases to distinguish:

        1. Without arguments::

               @decorator
               def myfunc(x):
                   pass

           which is equivalent to ::

               def myfunc(x):
                   pass

               myfunc = decorator(myfunc)


           Hence, in this case, the ``__new__`` method of the decorator
           immediately returns the wrapped function.

        2. With arguments::

               @decorator(*dec_args, **dec_kwargs)
               def myfunc(x):
                   pass

           which is equivalent to ::

               def myfunc(x):
                   pass

               dec_instance = decorator(*dec_args, **dec_kwargs)
               myfunc = dec_instance(myfunc)

           Hence, in this case, the first call creates an actual class
           instance of ``decorator``, and in the second statement, the
           ``dec_instance.__call__`` method returns the wrapper using
           the stored ``dec_args`` and ``dec_kwargs``.
        """
        # Decorating without arguments: return wrapper w/o args directly
        instance = super(OptionalArgDecorator, cls).__new__(cls)

        if (not kwargs and
                len(args) == 1 and
                callable(args[0])):
            func = args[0]

            return instance._wrapper(func)

        # With arguments, return class instance
        else:
            instance.wrapper_args = args
            instance.wrapper_kwargs = kwargs
            return instance

    def __call__(self, func):
        """Return ``self(func)``.

        This method is invoked when the decorator was created with
        arguments.

        Parameters
        ----------
        func : callable
            Original function to be wrapped

        Returns
        -------
        wrapped : callable
            The wrapped function
        """
        return self._wrapper(func, *self.wrapper_args, **self.wrapper_kwargs)

    @staticmethod
    def _wrapper(func, *wrapper_args, **wrapper_kwargs):
        """Make a wrapper for ``func`` and return it.

        This is a default implementation that simply returns the wrapped
        function, i.e., the resulting decorator is the identity.
        """
        return func


class vectorize(OptionalArgDecorator):

    """Decorator class for function vectorization.

    This vectorizer expects a function with exactly one positional
    argument (input) and optional keyword arguments. The decorated
    function has an optional ``out`` parameter for in-place evaluation.

    Examples
    --------
    Use the decorator witout arguments:

    >>> @vectorize
    ... def f(x):
    ...     return x[0] + x[1] if x[0] < x[1] else x[0] - x[1]
    >>>
    >>> f([0, 1])  # np.vectorize'd functions always return an array
    array(1)
    >>> f([[0, -2], [1, 4]])  # corresponds to points [0, 1], [-2, 4]
    array([1, 2])

    The function may have ``kwargs``:

    >>> @vectorize
    ... def f(x, param=1.0):
    ...     return x[0] + x[1] if x[0] < param else x[0] - x[1]
    >>>
    >>> f([[0, -2], [1, 4]])
    array([1, 2])
    >>> f([[0, -2], [1, 4]], param=-1.0)
    array([-1,  2])

    You can pass arguments to the vectorizer, too:

    >>> @vectorize(otypes=['float32'])
    ... def f(x):
    ...     return x[0] + x[1] if x[0] < x[1] else x[0] - x[1]
    >>> f([[0, -2], [1, 4]])
    array([ 1.,  2.], dtype=float32)
    """

    @staticmethod
    def _wrapper(func, *vect_args, **vect_kwargs):
        """Return the vectorized wrapper function."""
        if not hasattr(func, '__name__'):
            # Set name if not available. Happens if func is actually a function
            func.__name__ = '{}.__call__'.format(func.__class__.__name__)

        return wraps(func)(_NumpyVectorizeWrapper(func, *vect_args,
                                                  **vect_kwargs))


class _NumpyVectorizeWrapper(object):

    """Class for vectorization wrapping using `numpy.vectorize`.

    The purpose of this class is to store the vectorized version of
    a function when it is called for the first time.
    """

    def __init__(self, func, *vect_args, **vect_kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        func : callable
            Python function or method to be wrapped
        vect_args :
            positional arguments for `numpy.vectorize`
        vect_kwargs :
            keyword arguments for `numpy.vectorize`
        """
        super(_NumpyVectorizeWrapper, self).__init__()
        self.func = func
        self.vfunc = None
        self.vect_args = vect_args
        self.vect_kwargs = vect_kwargs

    def __call__(self, x, out=None, **kwargs):
        """Vectorized function call.

        Parameters
        ----------
        x : `array-like` or sequence of `array-like`'s
            Input argument(s) to the wrapped function
        out : `numpy.ndarray`, optional
            Appropriately sized array to write to

        Returns
        -------
        out : `numpy.ndarray`
            Result of the vectorized function evaluation. If ``out``
            was given, the returned object is a reference to it.
        """
        if np.isscalar(x):
            x = np.array([x])
        elif isinstance(x, np.ndarray) and x.ndim == 1:
            x = x[None, :]

        if self.vfunc is None:
            # Not yet vectorized
            def _func(*x, **kw):
                return self.func(np.array(x), **kw)

            self.vfunc = np.vectorize(_func, *self.vect_args,
                                      **self.vect_kwargs)

        if out is None:
            return self.vfunc(*x, **kwargs)
        else:
            out[:] = self.vfunc(*x, **kwargs)



if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division

import odl
from odl import vector
from odl.core.space.entry_points import TENSOR_SPACE_IMPLS
from odl.core.util.testutils import all_equal, default_precision_dict
import pytest 

error_dict = {
    'pytorch' : TypeError,
    'numpy'   : ValueError
}

def test_vector_numpy(odl_impl_device_pairs):

    impl, device = odl_impl_device_pairs
    tspace = TENSOR_SPACE_IMPLS[impl]((0))
    tspace_element_type = tspace.element_type

    inp = [[1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0]]

    x = vector(inp, impl=impl, device=device)

    assert isinstance(x, tspace_element_type)
    assert x.dtype_identifier == default_precision_dict[impl]['float']
    assert all_equal(x, inp)

    x = vector([1.0, 2.0, float('inf')], impl=impl, device=device)
    assert x.dtype_identifier == default_precision_dict[impl]['float']
    assert isinstance(x, tspace_element_type)

    x = vector([1.0, 2.0, float('nan')], impl=impl, device=device)
    assert x.dtype_identifier == default_precision_dict[impl]['float']
    assert isinstance(x, tspace_element_type)

    x = vector([1, 2, 3], dtype='float32', impl=impl, device=device)
    assert x.dtype_identifier == 'float32'
    assert isinstance(x, tspace_element_type)

    # Cn
    inp = [[1 + 1j, 2, 3 - 2j],
            [4 + 1j, 5, 6 - 1j]]

    x = vector(inp, impl=impl, device=device)
    assert isinstance(x, tspace_element_type)
    assert x.dtype_identifier == default_precision_dict[impl]['complex']
    assert all_equal(x, inp)

    x = vector([1, 2, 3], dtype='complex64', impl=impl, device=device)
    assert isinstance(x, tspace_element_type)

    # Generic TensorSpace
    inp = [1, 2, 3]
    x = vector(inp,impl=impl, device=device)
    assert isinstance(x, tspace_element_type)
    assert x.dtype_identifier == 'int64'
    assert all_equal(x, inp)

    inp = ['a', 'b', 'c']
    with pytest.raises(ValueError):
        x = vector(inp ,impl=impl, device=device)

    inp = [1, 2, 'inf']
    with pytest.raises(error_dict[impl]):
        x = vector(inp,impl=impl, device=device)

    # Scalar or empty input
    x = vector(5.0 ,impl=impl, device=device) # becomes 1d, size 1
    assert x.shape == ()

    x = vector([])  # becomes 1d, size 0
    assert x.shape == (0,)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import numpy as np
import pytest

import odl
from odl.core.discr.grid import sparse_meshgrid
from odl.core.util import is_int_dtype
from odl.core.util.testutils import all_equal
from odl.core.util.vectorization import (
    is_valid_input_array, is_valid_input_meshgrid,
    out_shape_from_meshgrid, out_shape_from_array,
    vectorize)


def test_is_valid_input_array():

    # 1d
    valid_shapes = [(1, 1), (1, 2), (1, 20), (20,)]
    invalid_shapes = [(2, 1), (1, 1, 1), (1,), ()]

    for shp in valid_shapes:
        arr = np.zeros(shp)
        assert is_valid_input_array(arr, ndim=1)

    for shp in invalid_shapes:
        arr = np.zeros(shp)
        assert not is_valid_input_array(arr, ndim=1)

    # 3d
    valid_shapes = [(3, 1), (3, 2), (3, 20)]
    invalid_shapes = [(3,), (20,), (4, 1), (3, 1, 1), ()]

    for shp in valid_shapes:
        arr = np.zeros(shp)
        assert is_valid_input_array(arr, ndim=3)

    for shp in invalid_shapes:
        arr = np.zeros(shp)
        assert not is_valid_input_array(arr, ndim=3)

    # Other input
    assert is_valid_input_array([[1, 2], [3, 4]], ndim=2)
    invalid_input = [1, [[[1, 2], [3, 4]]], (5,)]
    for inp in invalid_input:
        assert not is_valid_input_array(inp, ndim=2)


def test_is_valid_input_meshgrid():

    # 1d
    x = np.zeros(2)

    valid_mg = sparse_meshgrid(x)
    assert is_valid_input_meshgrid(valid_mg, ndim=1)

    invalid_mg = sparse_meshgrid(x, x)
    assert not is_valid_input_meshgrid(invalid_mg, ndim=1)

    x = np.zeros((2, 2))
    invalid_mg = sparse_meshgrid(x)
    assert not is_valid_input_meshgrid(invalid_mg, ndim=1)

    # 3d
    x, y, z = np.zeros(2), np.zeros(3), np.zeros(4)

    valid_mg = sparse_meshgrid(x, y, z)
    assert is_valid_input_meshgrid(valid_mg, ndim=3)

    invalid_mg = sparse_meshgrid(x, x, y, z)
    assert not is_valid_input_meshgrid(invalid_mg, ndim=3)

    x = np.zeros((3, 3))
    invalid_mg = sparse_meshgrid(x)
    assert not is_valid_input_meshgrid(invalid_mg, ndim=3)

    # Other input
    invalid_input = [1, [1, 2], ([1, 2], [3, 4]), (5,), np.zeros((2, 2))]
    for inp in invalid_input:
        assert not is_valid_input_meshgrid(inp, ndim=1)
        assert not is_valid_input_meshgrid(inp, ndim=2)


def test_out_shape_from_array():

    # 1d
    arr = np.zeros((1, 1))
    assert out_shape_from_array(arr) == (1,)
    arr = np.zeros((1, 2))
    assert out_shape_from_array(arr) == (2,)
    arr = np.zeros((1,))
    assert out_shape_from_array(arr) == (1,)
    arr = np.zeros((20,))
    assert out_shape_from_array(arr) == (20,)

    # 3d
    arr = np.zeros((3, 1))
    assert out_shape_from_array(arr) == (1,)
    arr = np.zeros((3, 2))
    assert out_shape_from_array(arr) == (2,)
    arr = np.zeros((3, 20))
    assert out_shape_from_array(arr) == (20,)


def test_out_shape_from_meshgrid():

    # 1d
    x = np.zeros(2)
    mg = sparse_meshgrid(x)
    assert out_shape_from_meshgrid(mg) == (2,)

    # 3d
    x, y, z = np.zeros(2), np.zeros(3), np.zeros(4)
    mg = sparse_meshgrid(x, y, z)
    assert out_shape_from_meshgrid(mg) == (2, 3, 4)

    # 3d, fleshed out meshgrids
    x, y, z = np.zeros(2), np.zeros(3), np.zeros(4)
    mg = np.meshgrid(x, y, z, sparse=False, indexing='ij', copy=True)
    assert out_shape_from_meshgrid(mg) == (2, 3, 4)

    mg = np.meshgrid(x, y, z, sparse=False, indexing='ij', copy=True)
    mg = tuple(reversed([np.asfortranarray(arr) for arr in mg]))
    assert out_shape_from_meshgrid(mg) == (2, 3, 4)


def test_vectorize_1d_otype():

    import sys

    # Test vectorization in 1d with given data type for output
    arr = (np.arange(5) - 2)[None, :]
    mg = sparse_meshgrid(np.arange(5) - 3)
    val_1 = -1
    val_2 = 2

    @vectorize(otypes=['int'])
    def simple_func(x):
        return 0 if x < 0 else 1

    true_result_arr = [0, 0, 1, 1, 1]
    true_result_mg = [0, 0, 0, 1, 1]

    # Out-of-place
    out = simple_func(arr)
    assert isinstance(out, np.ndarray)
    assert out.dtype == np.dtype('int')
    assert out.shape == (5,)
    assert all_equal(out, true_result_arr)

    out = simple_func(mg)
    assert isinstance(out, np.ndarray)
    assert out.shape == (5,)
    assert out.dtype == np.dtype('int')
    assert all_equal(out, true_result_mg)

    assert simple_func(val_1) == 0
    assert simple_func(val_2) == 1

    bogus_input = [lambda x: x, object, Exception]
    for b in bogus_input:
        with pytest.raises(TypeError):
            simple_func(b)

    # In-place
    out = np.empty(5, dtype='int')
    simple_func(arr, out=out)
    assert all_equal(out, true_result_arr)

    out = np.empty(5, dtype='int')
    simple_func(mg, out=out)
    assert all_equal(out, true_result_mg)


def test_vectorize_1d_lazy():

    # Test vectorization in 1d without data type --> lazy vectorization
    arr = (np.arange(5) - 2)[None, :]
    mg = sparse_meshgrid(np.arange(5) - 3)
    val_1 = -1
    val_2 = 2

    @vectorize
    def simple_func(x):
        return 0 if x < 0 else 1

    true_result_arr = [0, 0, 1, 1, 1]
    true_result_mg = [0, 0, 0, 1, 1]

    # Out-of-place
    out = simple_func(arr)
    assert isinstance(out, np.ndarray)
    assert is_int_dtype(out.dtype)
    assert out.shape == (5,)
    assert all_equal(out, true_result_arr)

    out = simple_func(mg)
    assert isinstance(out, np.ndarray)
    assert out.shape == (5,)
    assert is_int_dtype(out.dtype)
    assert all_equal(out, true_result_mg)

    assert simple_func(val_1) == 0
    assert simple_func(val_2) == 1


def test_vectorize_2d_dtype():

    # Test vectorization in 2d with given data type for output
    arr = np.empty((2, 5), dtype='int')
    arr[0] = ([-3, -2, -1, 0, 1])
    arr[1] = ([-1, 0, 1, 2, 3])
    mg = sparse_meshgrid([-3, -2, -1, 0, 1], [-1, 0, 1, 2, 3])
    val_1 = (-1, 1)
    val_2 = (2, 1)

    @vectorize(otypes=['int'])
    def simple_func(x):
        return 0 if x[0] < 0 and x[1] > 0 else 1

    true_result_arr = [1, 1, 0, 1, 1]

    true_result_mg = [[1, 1, 0, 0, 0],
                      [1, 1, 0, 0, 0],
                      [1, 1, 0, 0, 0],
                      [1, 1, 1, 1, 1],
                      [1, 1, 1, 1, 1]]

    # Out-of-place
    out = simple_func(arr)
    assert isinstance(out, np.ndarray)
    assert out.dtype == np.dtype('int')
    assert out.shape == (5,)
    assert all_equal(out, true_result_arr)

    out = simple_func(mg)
    assert isinstance(out, np.ndarray)
    assert out.dtype == np.dtype('int')
    assert out.shape == (5, 5)
    assert all_equal(out, true_result_mg)

    assert simple_func(val_1) == 0
    assert simple_func(val_2) == 1

    # In-place
    out = np.empty(5, dtype='int')
    simple_func(arr, out=out)
    assert all_equal(out, true_result_arr)

    out = np.empty((5, 5), dtype='int')
    simple_func(mg, out=out)
    assert all_equal(out, true_result_mg)


def test_vectorize_2d_lazy():

    # Test vectorization in 1d without data type --> lazy vectorization
    arr = np.empty((2, 5), dtype='int')
    arr[0] = ([-3, -2, -1, 0, 1])
    arr[1] = ([-1, 0, 1, 2, 3])
    mg = sparse_meshgrid([-3, -2, -1, 0, 1], [-1, 0, 1, 2, 3])
    val_1 = (-1, 1)
    val_2 = (2, 1)

    @vectorize
    def simple_func(x):
        return 0 if x[0] < 0 and x[1] > 0 else 1

    true_result_arr = [1, 1, 0, 1, 1]

    true_result_mg = [[1, 1, 0, 0, 0],
                      [1, 1, 0, 0, 0],
                      [1, 1, 0, 0, 0],
                      [1, 1, 1, 1, 1],
                      [1, 1, 1, 1, 1]]

    # Out-of-place
    out = simple_func(arr)
    assert isinstance(out, np.ndarray)
    assert is_int_dtype(out.dtype)
    assert out.shape == (5,)
    assert all_equal(out, true_result_arr)

    out = simple_func(mg)
    assert isinstance(out, np.ndarray)
    assert is_int_dtype(out.dtype)
    assert out.shape == (5, 5)
    assert all_equal(out, true_result_mg)

    assert simple_func(val_1) == 0
    assert simple_func(val_2) == 1


def test_vectorize_callable_class():

    # Test vectorization in 1d without data type --> lazy vectorization
    arr = [[-2, -1, 0, 1, 2]]
    mg = [[-3, -2, -1, 0, 1]]
    val_1 = -1
    val_2 = 2

    # Class with __call__ method
    class CallableClass(object):
        def __call__(self, x):
            return 0 if x < 0 else 1

    vectorized_call = vectorize(CallableClass())

    true_result_arr = [0, 0, 1, 1, 1]
    true_result_mg = [0, 0, 0, 1, 1]

    # Out-of-place
    out = vectorized_call(arr)
    assert isinstance(out, np.ndarray)
    assert is_int_dtype(out.dtype)
    assert out.shape == (5,)
    assert all_equal(out, true_result_arr)

    out = vectorized_call(mg)
    assert isinstance(out, np.ndarray)
    assert out.shape == (5,)
    assert is_int_dtype(out.dtype)
    assert all_equal(out, true_result_mg)

    assert vectorized_call(val_1) == 0
    assert vectorized_call(val_2) == 1


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Miscellaneous phantoms that do not fit in other categories."""

from __future__ import absolute_import, division, print_function

import sys

import numpy as np

__all__ = ('submarine', 'text')


def submarine(space, smooth=True, taper=20.0):
    """Return a 'submarine' phantom consisting in an ellipsoid and a box.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Discretized space in which the phantom is supposed to be created.
    smooth : bool, optional
        If ``True``, the boundaries are smoothed out. Otherwise, the
        function steps from 0 to 1 at the boundaries.
    taper : float, optional
        Tapering parameter for the boundary smoothing. Larger values
        mean faster taper, i.e. sharper boundaries.

    Returns
    -------
    phantom : ``space`` element
        The submarine phantom in ``space``.
    """
    if space.ndim == 2:
        if smooth:
            return _submarine_2d_smooth(space, taper)
        else:
            return _submarine_2d_nonsmooth(space)
    else:
        raise ValueError('phantom only defined in 2 dimensions, got {}'
                         ''.format(space.ndim))


def _submarine_2d_smooth(space, taper):
    """Return a 2d smooth 'submarine' phantom."""

    def logistic(x, c):
        """Smoothed step function from 0 to 1, centered at 0."""
        return 1. / (1 + np.exp(-c * x))

    def blurred_ellipse(x):
        """Blurred characteristic function of an ellipse.

        If ``space.domain`` is a rectangle ``[0, 1] x [0, 1]``,
        the ellipse is centered at ``(0.6, 0.3)`` and has half-axes
        ``(0.4, 0.14)``. For other domains, the values are scaled
        accordingly.
        """
        halfaxes = np.array([0.4, 0.14]) * space.domain.extent
        center = np.array([0.6, 0.3]) * space.domain.extent
        center += space.domain.min()

        # Efficiently calculate |z|^2, z = (x - center) / radii
        sq_ndist = np.zeros_like(x[0])
        for xi, rad, cen in zip(x, halfaxes, center):
            sq_ndist = sq_ndist + ((xi - cen) / rad) ** 2

        out = np.sqrt(sq_ndist)
        out -= 1
        # Return logistic(taper * (1 - |z|))
        return logistic(out, -taper)

    def blurred_rect(x):
        """Blurred characteristic function of a rectangle.

        If ``space.domain`` is a rectangle ``[0, 1] x [0, 1]``,
        the rect has lower left ``(0.56, 0.4)`` and upper right
        ``(0.76, 0.6)``. For other domains, the values are scaled
        accordingly.
        """
        xlower = np.array([0.56, 0.4]) * space.domain.extent
        xlower += space.domain.min()
        xupper = np.array([0.76, 0.6]) * space.domain.extent
        xupper += space.domain.min()

        out = np.ones_like(x[0])
        for xi, low, upp in zip(x, xlower, xupper):
            length = upp - low
            out = out * (logistic((xi - low) / length, taper) *
                         logistic((upp - xi) / length, taper))
        return out

    out = space.element(blurred_ellipse)
    out += space.element(blurred_rect)
    return out.ufuncs.minimum(1, out=out)


def _submarine_2d_nonsmooth(space):
    """Return a 2d nonsmooth 'submarine' phantom."""

    def ellipse(x):
        """Characteristic function of an ellipse.

        If ``space.domain`` is a rectangle ``[0, 1] x [0, 1]``,
        the ellipse is centered at ``(0.6, 0.3)`` and has half-axes
        ``(0.4, 0.14)``. For other domains, the values are scaled
        accordingly.
        """
        halfaxes = np.array([0.4, 0.14]) * space.domain.extent
        center = np.array([0.6, 0.3]) * space.domain.extent
        center += space.domain.min()

        sq_ndist = np.zeros_like(x[0])
        for xi, rad, cen in zip(x, halfaxes, center):
            sq_ndist = sq_ndist + ((xi - cen) / rad) ** 2

        return np.where(sq_ndist <= 1, 1, 0)

    def rect(x):
        """Characteristic function of a rectangle.

        If ``space.domain`` is a rectangle ``[0, 1] x [0, 1]``,
        the rect has lower left ``(0.56, 0.4)`` and upper right
        ``(0.76, 0.6)``. For other domains, the values are scaled
        accordingly.
        """
        xlower = np.array([0.56, 0.4]) * space.domain.extent
        xlower += space.domain.min()
        xupper = np.array([0.76, 0.6]) * space.domain.extent
        xupper += space.domain.min()

        out = np.ones_like(x[0])
        for xi, low, upp in zip(x, xlower, xupper):
            out = out * ((xi >= low) & (xi <= upp))
        return out

    out = space.element(ellipse)
    out += space.element(rect)
    return out.ufuncs.minimum(1, out=out)


def text(space, text, font=None, border=0.2, inverted=True):
    """Create phantom from text.

    The text is represented by a scalar image taking values in [0, 1].
    Depending on the choice of font, the text may or may not be anti-aliased.
    anti-aliased text can take any value between 0 and 1, while
    non-anti-aliased text produces a binary image.

    This method requires the ``pillow`` package.

    Parameters
    ----------
    space : `DiscretizedSpace`
        Discretized space in which the phantom is supposed to be created.
        Must be two-dimensional.
    text : str
        The text that should be written onto the background.
    font : str, optional
        The font that should be used to write the text. Available options are
        platform dependent.
        Default: Platform dependent. 'arial' for windows,
        'LiberationSans-Regular' for linux and 'Helvetica' for OSX
    border : float, optional
        Padding added around the text. 0.0 indicates that the phantom should
        occupy all of the space along its largest dimension while 1.0 gives a
        maximally padded image (text not visible).
    inverted : bool, optional
        If the phantom should be given in inverted style, i.e. white on black.

    Returns
    -------
    phantom : ``space`` element
        The text phantom in ``space``.

    Notes
    -----
    The set of available fonts is highly platform dependent, and there is no
    obvious way (except from trial and error) to find what fonts are supported
    on an arbitrary platform.

    In general, the fonts ``'arial'``, ``'calibri'`` and ``'impact'`` tend to
    be available on windows.

    Platform dependent tricks:

    **Linux**::

        $ find /usr/share/fonts -name "*.[to]tf"
    """
    from PIL import Image, ImageDraw, ImageFont

    if space.ndim != 2:
        raise ValueError('`space` must be two-dimensional')

    if font is None:
        platform = sys.platform
        if platform == 'win32':
            # Windows
            font = 'arial'
        elif platform == 'darwin':
            # Mac OSX
            font = 'Helvetica'
        else:
            # Assume platform is linux
            font = 'LiberationSans-Regular'

    text = str(text)

    # Figure out what font size we should use by creating a very high
    # resolution font and calculating the size of the text in this font
    init_size = 1000
    init_pil_font = ImageFont.truetype(font + ".ttf", size=init_size,
                                       encoding="unic")
    init_text_width, init_text_height = init_pil_font.getsize(text)

    # True size is given by how much too large (or small) the example was
    scaled_init_size = (1.0 - border) * init_size
    size = scaled_init_size * min([space.shape[0] / init_text_width,
                                   space.shape[1] / init_text_height])
    size = int(size)

    # Create font
    pil_font = ImageFont.truetype(font + ".ttf", size=size,
                                  encoding="unic")
    text_width, text_height = pil_font.getsize(text)

    # create a blank canvas with extra space between lines
    canvas = Image.new('RGB', space.shape, (255, 255, 255))

    # draw the text onto the canvas
    draw = ImageDraw.Draw(canvas)
    offset = ((space.shape[0] - text_width) // 2,
              (space.shape[1] - text_height) // 2)
    white = "#000000"
    draw.text(offset, text, font=pil_font, fill=white)

    # Convert the canvas into an array with values in [0, 1]
    arr = np.asarray(canvas)
    arr = np.sum(arr, -1)
    arr = arr / np.max(arr)
    arr = np.rot90(arr, -1)

    if inverted:
        arr = 1 - arr

    return space.element(arr)


if __name__ == '__main__':
    # Show the phantoms
    import odl
    from odl.core.util.testutils import run_doctests

    space = odl.uniform_discr([-1, -1], [1, 1], [300, 300])
    submarine(space, smooth=False).show('submarine smooth=False')
    submarine(space, smooth=True).show('submarine smooth=True')
    submarine(space, smooth=True, taper=50).show('submarine taper=50')

    text(space, text='phantom').show('phantom')

    # Run also the doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.
#
# The data is licensed under a
# Creative Commons Attribution 4.0 International License.
#
# You should have received a copy of the license along with this
# work. If not, see <http://creativecommons.org/licenses/by/4.0/>.

"""MRI datasets provided by TU Graz."""

import numpy as np

import odl
from odl.contrib.datasets.util import get_data

__all__ = ('mri_head_data_4_channel', 'mri_head_reco_op_4_channel',
           'mri_head_data_32_channel', 'mri_head_reco_op_32_channel',
           'mri_knee_data_8_channel', 'mri_knee_reco_op_8_channel')


DATA_SUBSET = 'trafos_tugraz'


def mri_head_data_4_channel():
    """Raw data for 4 channel MRI of a head.

    This is a T2 weighted TSE scan of a healthy volunteer.

    The data has been rescaled so that the reconstruction fits approximately
    in [0, 1].

    See the data source with DOI `10.5281/zenodo.800525`_ or the
    `project webpage`_ for further information.

    See Also
    --------
    mri_head_inverse_4_channel

    References
    ----------
    .. _10.5281/zenodo.800525: https://zenodo.org/record/800525
    .. _project webpage: http://imsc.uni-graz.at/mobis/internal/\
platform_aktuell.html
    """
    # TODO: Store data in some ODL controlled url
    url = 'https://zenodo.org/record/800525/files/1_rawdata_brainT2_4ch.mat'
    dct = get_data('1_rawdata_brainT2_4ch.mat', subset=DATA_SUBSET, url=url)

    # Change axes to match ODL definitions
    data = np.flip(np.swapaxes(dct['rawdata'], 0, -1) * 4e4, 2)

    return data


def mri_head_reco_op_4_channel():
    """Reconstruction operator for 4 channel MRI of a head.

    This is a T2 weighted TSE scan of a healthy volunteer.

    The reconstruction operator is the sum of the modulus of each channel.

    See the data source with DOI `10.5281/zenodo.800525`_ or the
    `project webpage`_ for further information.

    See Also
    --------
    mri_head_data_4_channel

    References
    ----------
    .. _10.5281/zenodo.800525: https://zenodo.org/record/800525
    .. _project webpage: http://imsc.uni-graz.at/mobis/internal/\
platform_aktuell.html
    """
    # To get the same rotation as in the reference article
    space = odl.uniform_discr(min_pt=[-115.2, -115.2],
                              max_pt=[115.2, 115.2],
                              shape=[256, 256],
                              dtype=complex)

    trafo = odl.trafos.FourierTransform(space)

    return odl.ReductionOperator(odl.ComplexModulus(space) * trafo.inverse, 4)


def mri_head_data_32_channel():
    """Raw data for 32 channel MRI of a head.

    This is a T2 weighted TSE scan of a healthy volunteer.

    The data has been rescaled so that the reconstruction fits approximately in
    [0, 1].

    See the data source with DOI `10.5281/zenodo.800527`_ or the
    `project webpage`_ for further information.

    See Also
    --------
    mri_head_inverse_32_channel

    References
    ----------
    .. _10.5281/zenodo.800527: https://zenodo.org/record/800527
    .. _project webpage: http://imsc.uni-graz.at/mobis/internal/\
platform_aktuell.html
    """
    # TODO: Store data in some ODL controlled url
    url = 'https://zenodo.org/record/800527/files/2_rawdata_brainT2_32ch.mat'
    dct = get_data('2_rawdata_brainT2_32ch.mat', subset=DATA_SUBSET, url=url)

    # Change axes to match ODL definitions
    data = np.flip(np.swapaxes(dct['rawdata'], 0, -1) * 7e3, 2)

    return data


def mri_head_reco_op_32_channel():
    """Reconstruction operator for 32 channel MRI of a head.

    This is a T2 weighted TSE scan of a healthy volunteer.

    The reconstruction operator is the sum of the modulus of each channel.

    See the data source with DOI `10.5281/zenodo.800527`_ or the
    `project webpage`_ for further information.

    See Also
    --------
    mri_head_data_32_channel

    References
    ----------
    .. _10.5281/zenodo.800529: https://zenodo.org/record/800527
    .. _project webpage: http://imsc.uni-graz.at/mobis/internal/\
platform_aktuell.html
    """
    # To get the same rotation as in the reference article
    space = odl.uniform_discr(min_pt=[-115.2, -115.2],
                              max_pt=[115.2, 115.2],
                              shape=[256, 256],
                              dtype=complex)

    trafo = odl.trafos.FourierTransform(space)

    return odl.ReductionOperator(odl.ComplexModulus(space) * trafo.inverse, 32)


def mri_knee_data_8_channel():
    """Raw data for 8 channel MRI of a knee.

    This is an SE measurement of the knee of a healthy volunteer.

    The data has been rescaled so that the reconstruction fits approximately in
    [0, 1].

    See the data source with DOI `10.5281/zenodo.800529`_ or the
    `project webpage`_ for further information.

    See Also
    --------
    mri_knee_inverse_8_channel

    References
    ----------
    .. _10.5281/zenodo.800529: https://zenodo.org/record/800529
    .. _project webpage: http://imsc.uni-graz.at/mobis/internal/\
platform_aktuell.html
    """
    # TODO: Store data in some ODL controlled url
    url = 'https://zenodo.org/record/800529/files/3_rawdata_knee_8ch.mat'
    dct = get_data('3_rawdata_knee_8ch.mat', subset=DATA_SUBSET, url=url)

    # Change axes to match ODL definitions
    data = np.flip(np.swapaxes(dct['rawdata'], 0, -1) * 9e3, 2)

    return data


def mri_knee_reco_op_8_channel():
    """Reconstruction operator for 8 channel MRI of a knee.

    This is an SE measurement of the knee of a healthy volunteer.

    The reconstruction operator is the sum of the modulus of each channel.

    See the data source with DOI `10.5281/zenodo.800529`_ or the
    `project webpage`_ for further information.

    See Also
    --------
    mri_knee_data_8_channel

    References
    ----------
    .. _10.5281/zenodo.800529: https://zenodo.org/record/800529
    .. _project webpage: http://imsc.uni-graz.at/mobis/internal/\
platform_aktuell.html
    """
    # To get the same rotation as in the reference article
    space = odl.uniform_discr(min_pt=[-74.88, -74.88],
                              max_pt=[74.88, 74.88],
                              shape=[192, 192],
                              dtype=complex)

    trafo = odl.trafos.FourierTransform(space)

    return odl.ReductionOperator(odl.ComplexModulus(space) * trafo.inverse, 8)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities mainly for internal use."""

from __future__ import absolute_import, division, print_function

import contextlib
from collections import OrderedDict
from contextlib import contextmanager
from itertools import product
from odl.core.util.print_utils import is_string
import numpy as np

__all__ = (
    'nd_iterator',
    'conj_exponent',
    'nullcontext',
    'writable_array',
    'run_from_ipython',
    'npy_random_seed',
    'unique',
)

def nd_iterator(shape):
    """Iterator over n-d cube with shape.

    Parameters
    ----------
    shape : sequence of int
        The number of points per axis

    Returns
    -------
    nd_iterator : generator
        Generator returning tuples of integers of length ``len(shape)``.

    Examples
    --------
    >>> for pt in nd_iterator([2, 2]):
    ...     print(pt)
    (0, 0)
    (0, 1)
    (1, 0)
    (1, 1)
    """
    return product(*map(range, shape))


def conj_exponent(exp):
    """Conjugate exponent ``exp / (exp - 1)``.

    Parameters
    ----------
    exp : positive float or inf
        Exponent for which to calculate the conjugate. Must be
        at least 1.0.

    Returns
    -------
    conj : positive float or inf
        Conjugate exponent. For ``exp=1``, return ``float('inf')``,
        for ``exp=float('inf')`` return 1. In all other cases, return
        ``exp / (exp - 1)``.
    """
    if exp == 1.0:
        return float('inf')
    elif exp == float('inf'):
        return 1.0
    else:
        return exp / (exp - 1.0)


@contextmanager
def nullcontext(enter_result=None):
    """Backport of the Python >=3.7 trivial context manager.

    See `the Python documentation
    <https://docs.python.org/3/library/contextlib.html#contextlib.nullcontext>`_
    for details.
    """
    try:
        yield enter_result
    finally:
        pass


try:
    nullcontext = contextlib.nullcontext
except AttributeError:
    pass


@contextmanager
def writable_array(obj, must_be_contiguous: bool =False):
    """Context manager that casts `obj` to a backend-specific array and saves changes
    made on that array back into `obj`.

    Parameters
    ----------
    obj : `array-like`
        Object that should be made available as writable array.
        It must be valid as input to `numpy.asarray` and needs to
        support the syntax ``obj[:] = arr``.
    must_be_contiguous : bool
        Whether the writable array should guarantee standard C order.

    Examples
    --------
    Usage with ODL vectors:

    >>> space = odl.uniform_discr(0, 1, 3)
    >>> x = space.element([1, 2, 3])
    >>> with writable_array(x) as arr:
    ...     arr += [1, 1, 1]
    >>> x
    uniform_discr(0.0, 1.0, 3).element([ 2.,  3.,  4.])

    Note that the changes are in general only saved upon exiting the 
    context manager. Before, the input object may remain unchanged.
    """
    if isinstance(obj, np.ndarray):
        if must_be_contiguous and not obj.data.c_contiguous:
            # Needs to convert to contiguous array
            arr = np.ascontiguousarray(obj)
            try:
                yield arr
            finally:
                obj[:] = arr
        else:
            try:
                yield obj
            finally:
                pass
    else:
        with obj.writable_array(must_be_contiguous=must_be_contiguous) as arr:
            yield arr

def run_from_ipython():
    """If the process is run from IPython."""
    return '__IPYTHON__' in globals()


def pkg_supports(feature, pkg_version, pkg_feat_dict):
    """Return bool indicating whether a package supports ``feature``.

    Parameters
    ----------
    feature : str
        Name of a potential feature of a package.
    pkg_version : str
        Version of the package that should be checked for presence of the
        feature.
    pkg_feat_dict : dict
        Specification of features of a package. Each item has the
        following form::

            feature_name: version_specification

        Here, ``feature_name`` is a string that is matched against
        ``feature``, and ``version_specification`` is a string or a
        sequence of strings that specifies version sets. These
        specifications are the same as for ``setuptools`` requirements,
        just without the package name.
        A ``None`` entry signals "no support in any version", i.e.,
        always ``False``.
        If a sequence of requirements are given, they are OR-ed together.
        See ``Examples`` for details.

    Returns
    -------
    supports : bool
        ``True`` if ``pkg_version`` of the package in question supports
        ``feature``, ``False`` otherwise.

    Examples
    --------
    >>> feat_dict = {
    ...     'feat1': '==0.5.1',
    ...     'feat2': '>0.6, <=0.9',  # both required simultaneously
    ...     'feat3': ['>0.6', '<=0.9'],  # only one required, i.e. always True
    ...     'feat4': ['==0.5.1', '>0.6, <=0.9'],
    ...     'feat5': None
    ... }
    >>> pkg_supports('feat1', '0.5.1', feat_dict)
    True
    >>> pkg_supports('feat1', '0.4', feat_dict)
    False
    >>> pkg_supports('feat2', '0.5.1', feat_dict)
    False
    >>> pkg_supports('feat2', '0.6.1', feat_dict)
    True
    >>> pkg_supports('feat2', '0.9', feat_dict)
    True
    >>> pkg_supports('feat2', '1.0', feat_dict)
    False
    >>> pkg_supports('feat3', '0.4', feat_dict)
    True
    >>> pkg_supports('feat3', '1.0', feat_dict)
    True
    >>> pkg_supports('feat4', '0.5.1', feat_dict)
    True
    >>> pkg_supports('feat4', '0.6', feat_dict)
    False
    >>> pkg_supports('feat4', '0.6.1', feat_dict)
    True
    >>> pkg_supports('feat4', '1.0', feat_dict)
    False
    >>> pkg_supports('feat5', '0.6.1', feat_dict)
    False
    >>> pkg_supports('feat5', '1.0', feat_dict)
    False
    """
    #This is an ugly workaround for the future deprecation of pkg_resources
    try:
        from pkg_resources import parse_requirements
    except ImportError:
        from packaging.requirements import Requirement
        def parse_requirements(s):
            return (Requirement(line) for line in s.splitlines() if line.strip() and not line.startswith("#"))

    feature = str(feature)
    pkg_version = str(pkg_version)
    supp_versions = pkg_feat_dict.get(feature, None)
    if supp_versions is None:
        return False

    # Make sequence from single string
    if is_string(supp_versions):
        supp_versions = [supp_versions]

    # Make valid package requirements
    ver_specs = ['pkg' + supp_ver for supp_ver in supp_versions]
    # Each parse_requirements list contains only one entry since we specify
    # only one package
    ver_reqs = [list(parse_requirements(ver_spec))[0]
                for ver_spec in ver_specs]

    # If one of the requirements in the list is met, return True
    for req in ver_reqs:
        if req.specifier.contains(pkg_version, prereleases=True):
            return True

    # No match
    return False


@contextmanager
def npy_random_seed(seed):
    """Context manager to temporarily set the NumPy random generator seed.

    Parameters
    ----------
    seed : int or None
        Seed value for the random number generator.
        ``None`` is interpreted as keeping the current seed.

    Examples
    --------
    Use this to make drawing pseudo-random numbers repeatable:

    >>> with npy_random_seed(42):
    ...     rand_int = np.random.randint(10)
    >>> with npy_random_seed(42):
    ...     same_rand_int = np.random.randint(10)
    >>> rand_int == same_rand_int
    True
    """
    do_seed = seed is not None
    orig_rng_state = None
    try:
        if do_seed:
            orig_rng_state = np.random.get_state()
            np.random.seed(seed)
        yield

    finally:
        if do_seed and orig_rng_state is not None:
            np.random.set_state(orig_rng_state)


def unique(seq):
    """Return the unique values in a sequence.

    Parameters
    ----------
    seq : sequence
        Sequence with (possibly duplicate) elements.

    Returns
    -------
    unique : list
        Unique elements of ``seq``.
        Order is guaranteed to be the same as in seq.

    Examples
    --------
    Determine unique elements in list

    >>> unique([1, 2, 3, 3])
    [1, 2, 3]

    >>> unique((1, 'str', 'str'))
    [1, 'str']

    The utility also works with unhashable types:

    >>> unique((1, [1], [1]))
    [1, [1]]
    """
    # First check if all elements are hashable, if so O(n) can be done
    try:
        return list(OrderedDict.fromkeys(seq))
    except TypeError:
        # Non-hashable, resort to O(n^2)
        unique_values = []
        for i in seq:
            if i not in unique_values:
                unique_values.append(i)
        return unique_values


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Primal-dual hybrid gradient (PDHG) algorithm studied by Chambolle and Pock.

The primal-dual hybrid gradient algorithm is a flexible method well suited for
non-smooth convex optimization problems in imaging.
"""

from __future__ import print_function, division, absolute_import
import numpy as np

from odl.core.operator import Operator


__all__ = ('pdhg', 'pdhg_stepsize')


# TODO: add dual gap as convergence measure
# TODO: diagonal preconditioning

def pdhg(x, f, g, L, niter, tau=None, sigma=None, **kwargs):
    r"""Primal-dual hybrid gradient algorithm for convex optimization.

    First order primal-dual hybrid-gradient method for non-smooth convex
    optimization problems with known saddle-point structure. The
    primal formulation of the general problem is ::

        min_{x in X} f(x) + g(L x)

    where ``L`` is an operator and ``f`` and ``g`` are functionals.

    The primal-dual hybrid-gradient algorithm is a primal-dual algorithm, and
    basically consists of alternating a gradient ascent in the dual variable
    and a gradient descent in the primal variable. The proximal operator is
    used to generate a ascent direction for the convex conjugate of F and
    descent direction for G. Additionally an over-relaxation of the primal
    variable is performed.

    Parameters
    ----------
    x : ``L.domain`` element
        Starting point of the iteration, updated in-place.
    f : `Functional`
        The function ``f`` in the problem definition. Needs to have
        ``f.proximal``.
    g : `Functional`
        The function ``g`` in the problem definition. Needs to have
        ``g.convex_conj.proximal``.
    L : linear `Operator`
        The linear operator that should be applied before ``g``. Its range must
        match the domain of ``g`` and its domain must match the domain of
        ``f``.
    niter : non-negative int
        Number of iterations.
    tau : float, optional
        Step size parameter for ``g``.
        Default: Sufficient for convergence, see `pdhg_stepsize`.
    sigma : sequence of floats, optional
        Step size parameters for ``f``.
        Default: Sufficient for convergence, see `pdhg_stepsize`.

    Other Parameters
    ----------------
    callback : callable, optional
        Function called with the current iterate after each iteration.
    theta : float, optional
        Relaxation parameter, required to fulfill ``0 <= theta <= 1``.
        Default: 1
    gamma_primal : non-negative float, optional
        Acceleration parameter. If not ``None``, it overrides ``theta`` and
        causes variable relaxation parameter and step sizes to be used,
        with ``tau`` and ``sigma`` as initial values. Requires ``f`` to be
        strongly convex and ``gamma_primal`` being upper bounded by the strong
        convexity constant of ``f``. Acceleration can either be done on the
        primal part or the dual part but not on both simultaneously.
        Default: ``None``
    gamma_dual : non-negative float, optional
        Acceleration parameter as ``gamma_primal`` but for dual variable.
        Requires ``g^*`` to be strongly convex and ``gamma_dual`` being upper
        bounded by the strong convexity constant of ``f^*``. Acceleration can
        either be done on the primal part or the dual part but not on both
        simultaneously.
        Default: ``None``
    x_relax : ``op.domain`` element, optional
        Required to resume iteration. For ``None``, a copy of the primal
        variable ``x`` is used.
        Default: ``None``
    y : ``op.range`` element, optional
        Required to resume iteration. For ``None``, ``op.range.zero()``
        is used.
        Default: ``None``

    Notes
    -----
    The problem of interest is

    .. math::
        \min_{x \in X} f(x) + g(L x),

    where the formal conditions are that :math:`L` is an operator
    between Hilbert spaces :math:`X` and :math:`Y`.
    Further, :math:`f : X \rightarrow [0, +\infty]` and
    :math:`g : Y \rightarrow [0, +\infty]` are proper, convex,
    lower-semicontinuous functionals.

    Convergence is only guaranteed if :math:`L` is linear, :math:`X, Y`
    are finite dimensional and the step lengths :math:`\sigma` and
    :math:`\tau` satisfy

    .. math::
       \tau \sigma \|L\|^2 < 1

    where :math:`\|L\|` is the operator norm of :math:`L`.

    It is often of interest to study problems that involve several operators,
    for example the classical TV regularized problem

    .. math::
        \min_x \|Ax - b\|_2^2 + \|\nabla x\|_1.

    Here it is tempting to let :math:`f(x)=\|\nabla x\|_1`, :math:`L=A` and
    :math:`g(y)=||y||_2^2`. This is however not feasible since the
    proximal of :math:`||\nabla x||_1` has no closed form expression.

    Instead, the problem can be formulated :math:`f(x)=0`,
    :math:`L(x) = (A(x), \nabla x)` and
    :math:`g((x_1, x_2)) = \|x_1\|_2^2 + \|x_2\|_1`. See the
    examples folder for more information on how to do this.

    For a more detailed documentation see `the PDHG guide
    <https://odlgroup.github.io/odl/guide/pdhg_guide.html>`_ in the online
    documentation.

    References on the algorithm can be found in `[CP2011a]
    <https://doi.org/10.1007/s10851-010-0251-1>`_ and `[CP2011b]
    <https://doi.org/10.1109/ICCV.2011.6126441>`_.

    This implementation of the CP algorithm is along the lines of
    `[Sid+2012] <https://doi.org/10.1088/0031-9155/57/10/3065>`_.

    The non-linear case is analyzed in `[Val2014]
    <https://doi.org/10.1088/0266-5611/30/5/055012>`_.

    See Also
    --------
    odl.solvers.nonsmooth.douglas_rachford.douglas_rachford_pd :
        Solver for similar problems which can additionaly handle infimal
        convolutions and multiple forward operators.
    odl.solvers.nonsmooth.forward_backward.forward_backward_pd :
        Solver for similar problems which can additionaly handle infimal
        convolutions, multiple forward operators and a differentiable term.

    References
    ----------
    [CP2011a] Chambolle, A and Pock, T. *A First-Order
    Primal-Dual Algorithm for Convex Problems with Applications to
    Imaging*. Journal of Mathematical Imaging and Vision, 40 (2011),
    pp 120-145.

    [CP2011b] Chambolle, A and Pock, T. *Diagonal
    preconditioning for first order primal-dual algorithms in convex
    optimization*. 2011 IEEE International Conference on Computer Vision
    (ICCV), 2011, pp 1762-1769.

    [Sid+2012] Sidky, E Y, Jorgensen, J H, and Pan, X.
    *Convex optimization problem prototyping for image reconstruction in
    computed tomography with the Chambolle-Pock algorithm*. Physics in
    Medicine and Biology, 57 (2012), pp 3065-3091.

    [Val2014] Valkonen, T.
    *A primal-dual hybrid gradient method for non-linear operators with
    applications to MRI*. Inverse Problems, 30 (2014).
    """
    # Forward operator
    if not isinstance(L, Operator):
        raise TypeError('`op` {!r} is not an `Operator` instance'
                        ''.format(L))

    # Starting point
    if x not in L.domain:
        raise TypeError('`x` {!r} is not in the domain of `op` {!r}'
                        ''.format(x, L.domain))

    # Spaces
    if f.domain != L.domain:
        raise TypeError('`f.domain` {!r} must equal `op.domain` {!r}'
                        ''.format(f.domain, L.domain))

    # Step size parameters
    tau, sigma = pdhg_stepsize(L, tau, sigma)

    # Number of iterations
    if not isinstance(niter, int) or niter < 0:
        raise ValueError('`niter` {} not understood'
                         ''.format(niter))

    # Relaxation parameter
    theta = kwargs.pop('theta', 1)
    theta, theta_in = float(theta), theta
    if not 0 <= theta <= 1:
        raise ValueError('`theta` {} not in [0, 1]'
                         ''.format(theta_in))

    # Acceleration parameters
    gamma_primal = kwargs.pop('gamma_primal', None)
    if gamma_primal is not None:
        gamma_primal, gamma_primal_in = float(gamma_primal), gamma_primal
        if gamma_primal < 0:
            raise ValueError('`gamma_primal` must be non-negative, got {}'
                             ''.format(gamma_primal_in))

    gamma_dual = kwargs.pop('gamma_dual', None)
    if gamma_dual is not None:
        gamma_dual, gamma_dual_in = float(gamma_dual), gamma_dual
        if gamma_dual < 0:
            raise ValueError('`gamma_dual` must be non-negative, got {}'
                             ''.format(gamma_dual_in))

    if gamma_primal is not None and gamma_dual is not None:
        raise ValueError('Only one acceleration parameter can be used')

    # Callback object
    callback = kwargs.pop('callback', None)
    if callback is not None and not callable(callback):
        raise TypeError('`callback` {} is not callable'
                        ''.format(callback))

    # Initialize the relaxation variable
    x_relax = kwargs.pop('x_relax', None)
    if x_relax is None:
        x_relax = x.copy()
    elif x_relax not in L.domain:
        raise TypeError('`x_relax` {} is not in the domain of '
                        '`L` {}'.format(x_relax.space, L.domain))

    # Initialize the dual variable
    y = kwargs.pop('y', None)
    if y is None:
        y = L.range.zero()
    elif y not in L.range:
        raise TypeError('`y` {} is not in the range of `L` '
                        '{}'.format(y.space, L.range))

    # Get the proximals
    proximal_primal = f.proximal
    proximal_dual = g.convex_conj.proximal
    proximal_constant = (gamma_primal is None) and (gamma_dual is None)
    if proximal_constant:
        # Pre-compute proximals for efficiency
        proximal_dual_sigma = proximal_dual(sigma)
        proximal_primal_tau = proximal_primal(tau)

    # Temporary copy to store previous iterate
    x_old = x.space.element()

    # Temporaries
    dual_tmp = L.range.element()
    primal_tmp = L.domain.element()

    for _ in range(niter):
        # Copy required for relaxation
        x_old.assign(x)

        # Gradient ascent in the dual variable y
        # Compute dual_tmp = y + sigma * L(x_relax)
        L(x_relax, out=dual_tmp)
        dual_tmp.lincomb(1, y, sigma, dual_tmp)

        # Apply the dual proximal
        if not proximal_constant:
            proximal_dual_sigma = proximal_dual(sigma)
        proximal_dual_sigma(dual_tmp, out=y)

        # Gradient descent in the primal variable x
        # Compute primal_tmp = x + (- tau) * L.derivative(x).adjoint(y)
        L.derivative(x).adjoint(y, out=primal_tmp)
        primal_tmp.lincomb(1, x, -tau, primal_tmp)

        # Apply the primal proximal
        if not proximal_constant:
            proximal_primal_tau = proximal_primal(tau)
        proximal_primal_tau(primal_tmp, out=x)

        # Acceleration
        if gamma_primal is not None:
            theta = float(1 / np.sqrt(1 + 2 * gamma_primal * tau))
            tau *= theta
            sigma /= theta

        if gamma_dual is not None:
            theta = float(1 / np.sqrt(1 + 2 * gamma_dual * sigma))
            tau /= theta
            sigma *= theta

        # Over-relaxation in the primal variable x
        x_relax.lincomb(1 + theta, x, -theta, x_old)

        if callback is not None:
            callback(x)


def pdhg_stepsize(L, tau=None, sigma=None):
    r"""Default step sizes for `pdhg`.

    Parameters
    ----------
    L : `Operator` or float
        Operator or norm of the operator that are used in the `pdhg` method.
        If it is an `Operator`, the norm is computed with
        ``Operator.norm(estimate=True)``.
    tau : positive float, optional
        Use this value for ``tau`` instead of computing it from the
        operator norms, see Notes.
    sigma : positive float, optional
        The ``sigma`` step size parameters for the dual update.

    Returns
    -------
    tau : float
        The ``tau`` step size parameter for the primal update.
    sigma : tuple of float
        The ``sigma`` step size parameter for the dual update.

    Notes
    -----
    To guarantee convergence, the parameters :math:`\tau`, :math:`\sigma`
    and :math:`L` need to satisfy

    .. math::
       \tau \sigma \|L\|^2 < 1

    This function has 4 options, :math:`\tau`/:math:`\sigma` given or not
    given.

    - Neither :math:`\tau` nor :math:`\sigma` are given, they are chosen as

      .. math::
          \tau = \sigma = \frac{\sqrt{0.9}}{\|L\|}

    - If only :math:`\sigma` is given, :math:`\tau` is set to

      .. math::
          \tau = \frac{0.9}{\sigma \|L\|^2}

    - If only :math:`\tau` is given, :math:`\sigma` is set
      to

      .. math::
          \sigma = \frac{0.9}{\tau \|L\|^2}

    - If both are given, they are returned as-is without further validation.
    """
    if tau is not None and sigma is not None:
        return float(tau), float(sigma)

    L_norm = L.norm(estimate=True) if isinstance(L, Operator) else float(L)
    if tau is None and sigma is None:
        tau = sigma = np.sqrt(0.9) / L_norm
        return tau, sigma
    elif tau is None:
        tau = 0.9 / (sigma * L_norm ** 2)
        return tau, float(sigma)
    else:  # sigma is None
        sigma = 0.9 / (tau * L_norm ** 2)
        return float(tau), sigma


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities for converting ODL operators to pytorch layers.

This requires the ``torch`` module from the ``pytorch`` package,
see `the pytorch installation guide
<https://github.com/pytorch/pytorch#installation>`_ for instructions.
"""

from __future__ import division

import warnings

import numpy as np
import torch
from packaging.version import parse as parse_version

from odl import Operator
from odl.core.space.base_tensors import Tensor, TensorSpace
from odl.core.util.npy_compat import AVOID_UNNECESSARY_COPY

from odl.backends.arrays.pytorch_tensors import pytorch_array_backend 

from typing import Optional


if parse_version(torch.__version__) < parse_version('0.4'):
    warnings.warn("This interface is designed to work with Pytorch >= 0.4",
                  RuntimeWarning, stacklevel=2)

__all__ = ('OperatorFunction', 'OperatorModule')


class OperatorFunction(torch.autograd.Function):
    """Wrapper of an ODL operator as a ``torch.autograd.Function``.

    This wrapper exposes an `Operator` object to the PyTorch autograd
    machinery by implementing custom ``forward()`` and ``backward()``
    methods.

    These methods should not be used directly. Instead, in a ``Module``,
    the call ``OperatorFunction.apply(operator, input_tensor)`` will
    apply the ``forward()`` method correctly and register gradients
    for the ``backward()`` step during backpropagation.

    The application of ``op`` to multiple inputs is done automatically
    in the background. The only requirement is that the shape of an
    input *ends with* the input shape that ``op`` expects, see below.

    Examples
    --------
    Simple example with of evaluating the ODL ``MatrixOperator`` on an
    input tensor of matching shape:

    >>> matrix = np.array([[1, 0, 1],
    ...                    [0, 1, 1]], dtype='float32')
    >>> odl_op = odl.MatrixOperator(matrix)
    >>> odl_op.domain.shape
    (3,)
    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> OperatorFunction.apply(odl_op, x)
    tensor([4., 5.])

    It is possible to pass tensors with extra axes "left" of the ones
    corresponding to the input shape expected by the operator:

    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> xs = x[None, None, :]  # shape (1, 1, 3)
    >>> OperatorFunction.apply(odl_op, xs)  # result shape (1, 1, 2)
    tensor([[[4., 5.]]])
    >>> xs = torch.stack([x, 2 * x], dim=0)  # shape (2, 3)
    >>> OperatorFunction.apply(odl_op, xs)  # result shape (2, 2)
    tensor([[ 4.,  5.],
            [ 8., 10.]])

    Functionals, i.e., operators with scalar output, are also supported:

    >>> odl_func = odl.functionals.L2NormSquared(odl.rn(3, dtype='float32'))
    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> OperatorFunction.apply(odl_func, x)
    tensor(14.)

    With multiple inputs:

    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> xs = torch.stack([x, 2 * x], dim=0)
    >>> OperatorFunction.apply(odl_func, xs)
    tensor([14., 56.])

    Backpropagation makes use of the Jacobian adjoint of the given matrix
    operator, which is transposed matrix operator. We mark the input
    tensor as requiring gradient, and compose the operator with the
    ``sum`` function to be able to backpropagate and get access to
    gradients:

    >>> x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    >>> loss = OperatorFunction.apply(odl_op, x).sum()
    >>> loss
    tensor(9., grad_fn=<SumBackward0>)
    >>> loss.backward()
    >>> x.grad  # should be matrix.T.dot([1, 1])
    tensor([1., 1., 2.])

    With multiple inputs:

    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> xs = torch.stack([x, 2 * x], dim=0).requires_grad_(True)
    >>> loss = OperatorFunction.apply(odl_op, xs).sum()
    >>> loss
    tensor(27., grad_fn=<SumBackward0>)
    >>> loss.backward()
    >>> xs.grad
    tensor([[1., 1., 2.],
            [1., 1., 2.]])

    We can again use a custom functional, with single or multiple
    inputs:

    >>> odl_func = odl.functionals.L2NormSquared(odl.rn(3, dtype='float32'))
    >>> x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    >>> loss = OperatorFunction.apply(odl_func, x)
    >>> loss
    tensor(14., grad_fn=<OperatorFunctionBackward>)
    >>> loss.backward()
    >>> x.grad  # should be 2 * x
    tensor([2., 4., 6.])
    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> xs = torch.stack([x, 2 * x], dim=0).requires_grad_(True)
    >>> loss = OperatorFunction.apply(odl_func, xs).sum()
    >>> loss
    tensor(70., grad_fn=<SumBackward0>)
    >>> loss.backward()
    >>> xs.grad  # should be 2 * xs
    tensor([[ 2.,  4.,  6.],
            [ 4.,  8., 12.]])

    Note, however, that the functional does not automatically reduce over
    extra axes, hence it cannot be used directly as a loss function. In
    addition, ODL functionals always take a single input.
    Loss functions of type ``loss_func(input, target)`` with reduction can
    be implemented e.g. as follows:

    >>> l2sq = odl.functionals.L2NormSquared(odl.rn(3, dtype='float32'))
    >>>
    >>> def my_mse(input, target, reduction='mean'):
    ...     val = OperatorFunction.apply(l2sq, input - target)
    ...     if reduction == 'mean':
    ...         return val.mean()
    ...     elif reduction == 'sum':
    ...         return val.sum()
    ...     elif reduction == 'none':
    ...         return val
    ...     else:
    ...         raise ValueError('bad reduction')
    ...
    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> xs = torch.stack([x, 2 * x], dim=0).requires_grad_(True)
    >>> ys = torch.stack([x, 2 * x], dim=0) + 1
    >>> loss = my_mse(xs, ys, reduction='sum')
    >>> loss
    tensor(6., grad_fn=<SumBackward0>)
    >>> loss.backward()
    >>> xs.grad
    tensor([[-2., -2., -2.],
            [-2., -2., -2.]])
    """

    @staticmethod
    def forward(ctx, operator: Operator, input_tensor: torch.Tensor) -> torch.Tensor:
        """Evaluate forward pass on the input.

        Parameters
        ----------
        ctx : context object
            Object to communicate information between forward and backward
            passes.
        operator : `Operator`
            ODL operator to be wrapped. For gradient computations to
            work, ``operator.derivative(x).adjoint`` must be implemented.
        input : `torch.Tensor`
            Point at which to evaluate the operator.

        Returns
        -------
        result : `torch.Tensor`
            Tensor holding the result of the evaluation.
        """
        assert(isinstance(input_tensor, torch.Tensor))
        assert(isinstance(operator, Operator))
        assert(isinstance(operator.domain, TensorSpace))
        ctx.operator = operator
        ctx.device = input_tensor.device

        input_tensor = input_tensor.detach()

        if not operator.is_linear:
            ctx.save_for_backward(input_tensor)

        input_arr = input_tensor.to(device=operator.domain.device)

        # Determine how to loop over extra shape "left" of the operator
        # domain shape
        in_shape = input_arr.shape
        op_in_shape = operator.domain.shape
        if operator.is_functional:
            op_out_shape = ()
            op_out_dtype = operator.domain.dtype
        else:
            op_out_shape = operator.range.shape
            op_out_dtype = operator.range.dtype

        extra_shape = in_shape[:-len(op_in_shape)]
        if in_shape[-len(op_in_shape):] != op_in_shape:
            shp_str = str(op_in_shape).strip('(,)')
            raise ValueError(
                'input tensor has wrong shape: expected (*, {}), got {}'
                ''.format(shp_str, in_shape)
            )

        # Store some information on the context object
        ctx.op_in_shape = op_in_shape
        ctx.op_out_shape = op_out_shape
        ctx.extra_shape = extra_shape
        ctx.op_in_dtype = operator.domain.dtype
        ctx.op_out_dtype = op_out_dtype

        def _apply_op_to_single_torch(single_input: torch.Tensor) -> torch.Tensor:
            x = operator.domain.element(single_input)
            y = operator(x)
            if isinstance(y, Tensor):
                y = pytorch_array_backend.from_dlpack(y.data)
            elif isinstance(y, (int, float, complex)):
                y = torch.tensor(y)
            else:
                raise TypeError(f"Unsupported result of type {type(y)} from operator.")
            return y.to(device=ctx.device)

        if extra_shape:
            input_arr_flat_extra = input_arr.reshape((-1,) + op_in_shape)
            results = []
            for inp in input_arr_flat_extra:
                results.append(_apply_op_to_single_torch(inp))

            # Stack results, reshape to the expected output shape and enforce
            # correct dtype
            result_arr = torch.stack(results)
            result = result_arr.reshape(extra_shape + op_out_shape)

        else:
            # Single input: evaluate directly
            result = _apply_op_to_single_torch(input_arr)

        return result



    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:
        r"""Apply the adjoint of the derivative at the input of the preceding
        ``forward`` call to ``grad_output``.

        This method is usually not called explicitly but as a part of the
        ``backward()`` pass of a backpropagation step.

        Parameters
        ----------
        ctx : context object
            Object to communicate information between forward and backward
            passes.
        grad_output : `torch.Tensor`
            Tensor to which the Jacobian should be applied. See Notes
            for details.

        Returns
        -------
        gradients : tuple
            Tuple ``(None, grad_input)``, where the ``None`` part is due to
            the first argument of ``forward`` being the ODL operator that
            does not require a gradient. The ``grad_input`` tensor is the
            result of applying the Jacobian to ``grad_output``.
            See Notes for details.

        Notes
        -----
        This method applies the contribution of this node, i.e., the
        transpose of the Jacobian of its outputs with respect to its inputs,
        to the gradients of some cost function with respect to the outputs
        of this node.

        **Example:** Assume that this node computes :math:`x \mapsto C(f(x))`,
        where :math:`x` is a tensor and :math:`C` is a scalar-valued
        function. In ODL language, what ``backward`` should compute is

        .. math::
            \nabla(C \circ f)(x) = f'(x)^*\big(\nabla C (f(x))\big)

        according to the chain rule. In ODL code, this corresponds to ::

            f.derivative(x).adjoint(C.gradient(f(x))).

        Hence, the parameter ``grad_output`` is a tensor containing
        :math:`y = \nabla C(f(x))`. Then, ``backward`` boils down to
        computing ``[f'(x)^*(y)]`` using the input ``x`` stored during
        the previous `forward` pass.
        """

        # Return early if there's nothing to do
        if not ctx.needs_input_grad[1]:
            return None, None

        operator = ctx.operator

        try:
            dom_weight = operator.domain.weighting.const
        except AttributeError:
            dom_weight = 1.0
        try:
            ran_weight = operator.range.weighting.const
        except AttributeError:
            ran_weight = 1.0
        scaling = dom_weight / ran_weight

        grad_output_arr = grad_output.to(device=operator.domain.device)

        op_in_shape = ctx.op_in_shape
        op_out_shape = ctx.op_out_shape
        extra_shape = ctx.extra_shape

        def _apply_op_to_single_torch( single_input: Optional[torch.Tensor]
                                     , single_grad_out: torch.Tensor ) -> torch.Tensor:
            g = operator.range.element(single_grad_out)
            if operator.is_linear:
                result = operator.adjoint(g)
            else:
                x = operator.domain.element(single_input)
                result = operator.derivative(x).adjoint(g)
            return pytorch_array_backend.from_dlpack(result.data).to(ctx.device)

        if not operator.is_linear:
            input_arr = ctx.saved_tensors[0].detach()

        if extra_shape:
            # Multiple gradients: flatten extra axes, then do one entry
            # at a time
            grad_output_arr_flat_extra = grad_output_arr.reshape(
                (-1,) + op_out_shape
            )

            results = []

            if operator.is_linear:
                for ograd in grad_output_arr_flat_extra:
                    results.append(_apply_op_to_single_torch(None, ograd))
            else:
                # Need inputs, flattened in the same way as the gradients
                input_arr_flat_extra = input_arr.reshape((-1,) + op_in_shape)
                for ograd, inp in zip(grad_output_arr_flat_extra, input_arr_flat_extra):
                    results.append(_apply_op_to_single_torch(inp, ograd))

            # Stack results, reshape to the expected output shape and enforce
            # correct dtype
            result_tensor = torch.stack(results).reshape(extra_shape + op_in_shape)
        else:
            if operator.is_linear:
                result_tensor = _apply_op_to_single_torch(None, grad_output.detach())
            else:
                result_tensor = _apply_op_to_single_torch(input_arr, grad_output.detach())
        
        if scaling != 1.0:
            result_tensor *= scaling

        return None, result_tensor



        



class OperatorModule(torch.nn.Module):

    """Wrapper of an ODL operator as a ``torch.nn.Module``.

    This wrapper can be used as a layer in ``pytorch`` Neural Networks.
    It works with arbitrary batches and channels and supports
    backpropagation.

    Parameters
    ----------
    operator : `Operator`
        The ODL operator to be wrapped. For gradient computations to work,
        ``operator.derivative(x).adjoint`` must be implemented.

    Examples
    --------
    Simple example of using wrapping a ``MatrixOperator`` as a ``Module``.
    The input must have at least one extra dimension (batch axis), i.e.,
    in this case must be a 2D tensor:

    >>> matrix = np.array([[1, 0, 0],
    ...                    [0, 1, 1]], dtype='float32')
    >>> odl_op = odl.MatrixOperator(matrix)
    >>> odl_op.domain.shape
    (3,)
    >>> odl_op.range.shape
    (2,)
    >>> op_mod = OperatorModule(odl_op)
    >>> x = torch.ones((1, 3))  # with trivial batch axis
    >>> op_mod(x)
    tensor([[1., 2.]])
    >>> t = torch.ones(3)
    >>> x = torch.stack([0 * t, 1 * t])  # batch size 2
    >>> op_mod(x)
    tensor([[0., 0.],
            [1., 2.]])

    An arbitrary number of axes is supported:

    >>> x = t[None, None, :]  # trivial batch and channel
    >>> op_mod(x)
    tensor([[[1., 2.]]])
    >>> x = torch.stack([torch.stack([0 * t, 1 * t]),
    ...                  torch.stack([2 * t, 3 * t]),
    ...                  torch.stack([4 * t, 5 * t])])
    >>> op_mod(x)
    tensor([[[ 0.,  0.],
             [ 1.,  2.]],
    <BLANKLINE>
            [[ 2.,  4.],
             [ 3.,  6.]],
    <BLANKLINE>
            [[ 4.,  8.],
             [ 5., 10.]]])

    Backpropagation works autmatically by means of the
    ``operator.derivative(x).adjoint`` machinery. To trigger it, the
    input tensor must be marked as requiring gradient:

    >>> x = torch.tensor([[1.0, 2.0, 3.0]], requires_grad=True)
    >>> loss = op_mod(x).sum()
    >>> loss
    tensor(6., grad_fn=<SumBackward0>)
    >>> loss.backward()
    >>> x.grad
    tensor([[1., 1., 1.]])
    """

    def __init__(self, operator):
        """Initialize a new instance."""
        super(OperatorModule, self).__init__()
        self.operator = operator

    def forward(self, x):
        """Compute forward-pass of this module on ``x``.

        Parameters
        ----------
        x : `torch.Tensor`
            Input of this layer. The contained tensor must have shape
            ``extra_shape + operator.domain.shape``, and
            ``len(extra_shape)`` must be at least 1 (batch axis).

        Returns
        -------
        out : `torch.Tensor`
            The computed output. Its tensor will have shape
            ``extra_shape + operator.range.shape``, where ``extra_shape``
            are the extra axes of ``x``.

        Examples
        --------
        """
        in_shape = tuple(x.shape)
        in_ndim = len(in_shape)
        op_in_shape = self.operator.domain.shape
        op_in_ndim = len(op_in_shape)
        if in_ndim <= op_in_ndim or in_shape[-op_in_ndim:] != op_in_shape:
            shp_str = str(op_in_shape).strip('()')
            raise ValueError(
                'input tensor has wrong shape: expected (N, *, {}), got {}'
                ''.format(shp_str, in_shape)
            )
        return OperatorFunction.apply(self.operator, x)

    def __repr__(self):
        """Return ``repr(self)``."""
        op_name = self.operator.__class__.__name__
        op_in_shape = self.operator.domain.shape
        if len(op_in_shape) == 1:
            op_in_shape = op_in_shape[0]
        op_out_shape = self.operator.range.shape
        if len(op_out_shape) == 1:
            op_out_shape = op_out_shape[0]

        return '{}({}) ({} -> {})'.format(
            self.__class__.__name__, op_name, op_in_shape, op_out_shape
        )


def copy_if_zero_strides(arr):
    """Workaround for NumPy issue #9165 with 0 in arr.strides."""
    assert isinstance(arr, np.ndarray)
    return arr.copy() if 0 in arr.strides else arr


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    import odl
    from torch import autograd, nn
    run_doctests(extraglobs={'np': np, 'odl': odl, 'torch': torch,
                             'nn': nn, 'autograd': autograd})



import pytest

import odl

from odl.core.util.pytest_config import IMPL_DEVICE_PAIRS
from odl.core.util.testutils import (
    noise_elements, simple_fixture)

DEFAULT_SHAPE = (4,4)

elementwise_comparison = simple_fixture(
    'elementwise', ["isclose" ]
    )

reduction_comparison = simple_fixture(
    'reduction', ["allclose", "odl_all_equal"]
    )

truth_value_comparison = simple_fixture(
    'truth_value', ["all", "any",]
    )


@pytest.fixture(scope='module', params=IMPL_DEVICE_PAIRS)
def float_tspace(request, odl_real_floating_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_real_floating_dtype, 
        impl=impl, 
        device=device
    )

def test_elementwise(float_tspace, elementwise_comparison):
    ns = float_tspace.array_namespace
    arr_fn = getattr(ns, elementwise_comparison)
    odl_fn = getattr(odl, elementwise_comparison)

    xarr0, x0 = noise_elements(float_tspace, 1)
    xarr1, x1 = noise_elements(float_tspace, 1)

    assert (arr_fn(xarr0, xarr0) == odl_fn(x0, x0)).all()
    assert (arr_fn(xarr0, xarr1) == odl_fn(x0, x1)).all()
    assert (arr_fn(xarr1, xarr0) == odl_fn(x1, x0)).all()

def test_reduction(float_tspace, reduction_comparison):
    ns = float_tspace.array_namespace
    xarr0, x0 = noise_elements(float_tspace, 1)
    xarr1, x1 = noise_elements(float_tspace, 1)
    odl_fn = getattr(odl, reduction_comparison)

    if reduction_comparison == 'allclose':
        arr_fn = getattr(ns, reduction_comparison)
        
    elif reduction_comparison == 'odl_all_equal':
        all_fn   = getattr(ns, 'all')
        equal_fn = getattr(ns, 'equal')
        def arr_fn(x, y):
            return all_fn(equal_fn(x, y)) 
        
    else:
        raise ValueError
        
    assert arr_fn(xarr0, xarr0) == odl_fn(x0, x0)
    assert arr_fn(xarr0, xarr1) == odl_fn(x0, x1)
    assert arr_fn(xarr1, xarr0) == odl_fn(x1, x0)

def test_array_truth_value(float_tspace, truth_value_comparison):
    ns = float_tspace.array_namespace
    arr_fn = getattr(ns, truth_value_comparison)
    odl_fn = getattr(odl, truth_value_comparison)

    xarr0, x0 = noise_elements(float_tspace, 1)
    xarr1, x1 = noise_elements(float_tspace, 1)

    arr_isclose = getattr(ns, 'isclose')
    odl_isclose = getattr(odl, 'isclose')

    expr_0 = arr_isclose(xarr0, xarr0) == odl_isclose(x0, x0)
    expr_1 = arr_isclose(xarr0, xarr1) == odl_isclose(x0, x1)
    expr_2 = arr_isclose(xarr1, xarr0) == odl_isclose(x1, x0)
    assert arr_fn(expr_0) == odl_fn(expr_0)
    assert arr_fn(expr_1) == odl_fn(expr_1)
    assert arr_fn(expr_2) == odl_fn(expr_2)
    
